import time
import numpy as np
from tlux.approximate.balltree import BallTree
from sklearn.neighbors import KDTree
# from sklearn.neighbors import BallTree

print("Initializing data..")
np.random.seed(1)
x = np.random.random(size=(10000000, 32))
x[x.shape[0]//2:,:] = x[:x.shape[0]//2,:]
i = np.random.randint(x.shape[0])
print()

print("Building tree..")
time.sleep(0.1)
# s = time.time(); t = KDTree(x); e = time.time(); print(f"{e-s:.6f} second fit")
s = time.time(); t = BallTree(x); e = time.time(); print(f"{e-s:.6f} second fit")
print()

time.sleep(0.1)
s = time.time(); print(t.query(x[i:i+1])[0][0].tolist()); e = time.time()
print(f"{e-s:.6f} seconds (1 nearest)")
print()

time.sleep(0.1)
s = time.time(); print(t.query(x[i:i+1], k=5)[0][0].tolist()); e = time.time()
print(f"{e-s:.6f} seconds (5 nearest)")
print()

time.sleep(0.1)
s = time.time(); print(t.query(x[i:i+1], k=100)[0][0].tolist()); e = time.time()
print(f"{e-s:.6f} seconds (100 nearest)")
print()
print()

try:
    time.sleep(0.1)
    s = time.time(); print(t.query(x[i:i+1], k=5, budget=100000)[0][0].tolist()); e = time.time();
    print(f"{e-s:.6f} seconds (100K budget)")
    print()

    time.sleep(0.1)
    s = time.time(); print(t.query(x[i:i+1], k=5, budget=100)[0][0].tolist()); e = time.time();
    print(f"{e-s:.6f} seconds (100 budget)")
    print()
    print()

    for i in range(10):
        time.sleep(0.01)
        s = time.time(); print(t.query(x[i:i+1], k=5, budget=10000, randomness=0.2)[0][0].tolist()); e = time.time();
        print(f"{e-s:.6f} seconds (10K budget)")
        print()
except: pass




import numpy as np
from tlux.approximate.base import Approximator

# Wrapper class for using the Delaunay scipy code.
class Delaunay(Approximator):
    def __init__(self):
        from scipy.spatial import Delaunay as qHullDelaunay
        self.qHullDelaunay = qHullDelaunay
        self.x = None
        self.y = None

    # Use fortran code to compute the boxes for the given data
    def _fit(self, x, y):
        # Delaunay is only made for more than 1 dimension.
        if (x.shape[1] > 1):
            self.mesh = self.qHullDelaunay(x)
        else:
            # Define a 1D function that linearly interpolates the data.
            x = x.flatten()
            i = np.argsort(x)
            x = x[i]
            y = y[i]
            gaps = x[1:] - x[:-1]
            def interpolate(z, x=x, y=y):
                v = []
                for p in z.flatten():
                    if   (p <= x[0]):  v.append(y[0])
                    elif (p >= x[-1]): v.append(y[-1])
                    else:
                        i = np.searchsorted(x, p, side="right")
                        rval = (p - x[i-1])
                        lval = (x[i] - p)
                        v.append( (lval*y[i-1] + rval*y[i]) / (lval+rval) )
                v = np.asarray(v)
                return v
            # Store this 1D function.
            self.mesh = interpolate
        # Store the x and y values for prediction.
        self.x = x
        self.y = y
        self.y_min = y.min(axis=0)
        self.y_max = y.max(axis=0)


    # Function that returns the indices of points and the weights that
    # should be used to make associated predictions for each point in
    # "points".
    def _predict(self, x):
        if (len(self.x.shape) > 1):
            # Solve for the weights in the Delaunay model
            simp_ind = self.mesh.find_simplex(x)
            # If a point is outside the convex hull, use the
            # closest simplex to extrapolate the value
            simp_ind = np.where(
                simp_ind == -1,
                np.argmax(self.mesh.plane_distance(x), axis=1),
                simp_ind
            )
            simp = self.mesh.simplices[simp_ind]
            zx = self.x[simp.flatten()].reshape(simp.shape+(self.x.shape[-1],))
            # Solve for the response value with a linear solve.
            system = np.concatenate((zx,np.ones(zx.shape[:-1]+(1,))), axis=-1)
            system = np.transpose(system, axes=(0,2,1))
            x_pts = np.concatenate((x,np.ones(x.shape[:-1]+(1,))), axis=-1)
            weights = np.linalg.solve(system, x_pts)
            # Get the y values.
            zy = self.y[simp.flatten()].reshape(simp.shape+(self.y.shape[-1],))
            y = np.zeros((x.shape[0], zy.shape[-1]))
            for i in range(x.shape[0]):
                y[i,:] = np.matmul(zy[i].T, weights[i,:])
            # Clip to not leave range of observed values.
            return np.clip(y, self.y_min, self.y_max)
        else:
            return self.mesh(x)


# Define a simple test to make sure it is working.
if __name__ == "__main__":
    from tlux.plot import Plot
    from tlux.random import well_spaced_box, well_spaced_ball

    # Generate some data.
    x = well_spaced_box(400, 2)
    y = np.cos(10*np.linalg.norm(x, axis=1))

    # Fit a nearest neighbor model.
    m = Delaunay()
    m.fit(x, y)

    # Create a visual.
    p = Plot()
    p.add("data", *x.T, y)
    p.add_func("fit", m, *([[0,1]]*x.shape[1]), vectorized=True)
    p.show(z_range=[-2,4])




from tlux.data.data import Data, flatten, merge_on_column

# TODO:  Test for file that has a type digression happen on a line
#        with empty strings in it. This caused a crash [2019-12-17].
# TODO:  Test data with no names getting other data added in place,
#        should overwrite the names!


# Some tests for the data
def test_data():
    import os
    import numpy as np

    print("Testing Data...", end=" ")

    # Set the max wait to zero and the file to a null output
    #  (to instigate all loading print statements without showing).
    _mw = Data.max_wait
    _pf = Data.print_kwargs.get("file",None)
    Data.max_wait = 0
    Data.print_kwargs["file"] = open(os.devnull,"w")

    # ----------------------------------------------------------------
    a = Data()

    # Verify copy of empty data.
    b = a.copy()

    # Verify append
    a.append([1,"a"])
    a.append([2,"b"])
    a.append([3,"c"])

    # Verify add column (with edge case, none type)
    a.add_column([None,None,None])
    assert(a.types[-1] == type(None))

    # Reassign the missing value column to floats, verify type update
    a[a.names[-1]] = [1.2, 3.0, 2.4]
    assert(a.types[-1] == float)
    # WARNING: Need to check doing (<int>, <str>) assignment

    # Verify automatic type generation AND automatic type-casting
    a.append([1,2,3])
    assert(tuple(a[-1]) == tuple([1,"2",3.0]))
    assert(tuple(map(type,a[-1])) == (int, str, float))

    # Add some missing values.
    a.append([-1,None,None])

    #  Done modifying "a", rest of tests leave it constant.
    # ----------------------------------------------------------------

    # Verify that "index" works correctly.
    assert( a.index([1,"2",3.0]) == 3 )
    assert( a.index([-1,None,None]) == 4 )

    # Verify that slicing works on descirptors.
    assert(a.names[:-1] == ['0','1'])

    # Verify that slicing works on descriptors of views.
    b = a[:,1:]
    assert(b.names[1:] == ['2'])

    # Verify in-place addition of rows (different length data with same column names)
    b = a[:-2].copy()
    c = a[1:-2]
    b += c
    assert(tuple(b["0"]) == tuple([1,2,3,2,3]))

    # Verify in-place addition of new columns (same length data with new columns)
    b = a[:,:-1].copy()
    c = a[:,-1:].copy()
    c["3"] = range(len(c))
    b += c
    assert(tuple(b[-2]) == (1,"2",3.0,3))
    assert(tuple(b[-1]) == (-1,None,None,4))

    # Verify in-place addition of new columns (same length data with fewer columns)
    b = a[:,:-1].copy()
    c = a[:,:].copy()
    c["3"] = range(len(c))
    c += b
    assert(tuple(c[-2]) == (1,"2",None,None))
    assert(tuple(c[:,-1]) == tuple(range(5)) + (None,)*5)

    # Verify the addition of in place addition of new columns AND rows.
    b = a[:]
    b['3'] = range(len(b))
    c = a[:]
    c += b
    assert(tuple(c['3']) == tuple([None]*len(b) + list(range(len(c)-len(b)))))

    # Verify in-place addition of a data set with *some* duplicate
    # column names, but the same number of rows (adds columns).
    b = a[:,1:].copy()
    c = a[:,:-1].copy()
    c.names = range(3,3+c.shape[1])
    b += c
    assert(b.shape == (5,4))
    assert(tuple(b.names) == tuple(map(str,range(1,5))))
    assert(len(list(b[:,-1] == c[:,-1])) == b.shape[0])

    # Verify slicing-based singleton value assignment
    b = a[:-2].copy()
    b[:,0] = 1
    assert(tuple(b["0"]) == tuple([1,1,1]))

    # Verify slicing-based multiple value assignment
    b = a[:-2].copy()
    b[:,0] = [3,1,4]
    assert(tuple(b["0"]) == tuple([3,1,4]))

    # Verify double indexing with integers
    assert(a[0,1] == "a")

    # Verify double indexing with slices
    assert(tuple(a[::-1,1])[2:] == tuple(["c","b","a"]))

    # Verify standard index access
    assert(tuple(a[0]) == tuple([1,"a",1.2]))

    # Verify column-access and automatic column naming
    assert(tuple(a["0"])[:-2] == tuple([1,2,3]))

    # Verify slicing by index
    assert(tuple(a[:2][0]) == tuple(a[0]))

    # Verify slicing by names
    assert(tuple(a["0","2"][0]) == tuple([1,1.2]))

    # Verify that copies with "copy" are deep
    b = a.copy()
    b.retype([str,str,str])
    assert(a.types[0] != str)

    # Assert that the two forms of full-data slicing make copies.
    assert(id(a[:].data) != id(a.data))
    assert(id(a[:,:].data) != id(a.data))

    # Verify that copies with slicing are deep and that retype works.
    b = a[:]
    b.retype([str,str,str])
    assert(a.types[0] != str)

    # Verify that in-place addition of data with same names in
    # different order is handled correctly.
    b = a[:,:]
    c = a[:,['1','0','2']]
    b += c
    assert(tuple(b[:,1]) == 2*tuple(a[:,1]))

    # Verify that multiple transformations on data yield correct view.
    b = a[:]
    b += a
    b += a
    c = b[::2]
    c = c[['1', '2']].unique()
    assert(c.shape[0] == len(a))
    assert(tuple(c.names) == tuple(b.names[1:]))

    # Verify that sorting a data object and a view object both work.
    b = a[:]
    b.sort()
    assert(tuple(b[:,0]) == tuple(sorted(a[:,0])))
    b = a[:-1, [0,1]]
    b.sort()
    assert(tuple(b[:,1]) == tuple(sorted(a[:-1,1])))
    # Assert that "a" (the object being viewed) is unmodified.
    assert(tuple(a['1']) == ('a','b','c','2',None))

    # Verify the stack process.
    b = a[:]
    c = a[:]
    c['2'] += 1.0
    b += c
    b.stack('2')
    assert(tuple(b[0,-1]) == (a[0,-1], a[0,-1]+1.0))

    # Testing the "unstack" operation.
    b = a[:]
    b += a
    b.pop(-1)
    b.stack('2')
    # Convert the "stacked" column into generator ogbjects.
    b['2'] = ((v for v in row) for row in b['2'])
    b[-1,'2'] = None
    # Unstack, and verify that the expected output happens.
    b.unstack('2')
    assert(tuple(b['1']) == ('a','a','b','b','c','c','2','2',None))

    # Verify that reorder works.
    b = a[:]
    b.reorder(["2","0"])
    assert(tuple(b.names) == ("2","0","1"))
    assert(tuple(b.types) == (float,int,str))
    for i in range(len(b)):
        for n in b.names:
            assert(a[i,n] == b[i,n])

    # Verify the ability to construct real-valued arrays (and go back)
    numeric = a.to_matrix()
    c = Data(map(numeric.from_real, numeric))
    assert(tuple(a[:-1]["1"]) == tuple(c[:-1]["1"]))

    # Verify column item retrieval and assignment
    assert(a["0"][0] == a[0][0])
    a["0"][0] = -10
    a["0"][0] = 1

    # Verify total column assignment
    first_col = list(a["0"])
    new_col = list(map(str,a["0"]))
    a["0"] = new_col
    assert(tuple(a["0"]) == tuple(new_col))
    a["0"] = first_col

    # Verify new column assignment
    b = a[:]
    first_col = list(b["0"])
    new_col = list(map(str,b["0"]))
    b["0-str"] = new_col
    assert(tuple(map(str,b["0"])) == tuple(b["0-str"]))

    # Verify new column assignment for empty Data.
    b = Data()
    n = a.names[0]
    b[n] = a[n]
    assert(all(b[n] == a[n]))

    # Verify that multiple column assignment for empty Data fails (need tests otherwise).
    b = Data()
    n = list(a.names)
    try: b[n] = a[n]
    except Data.Unsupported: pass
    else: assert(False)

    # Verify copying a data object that has a 'None' in the first row
    b = Data(names=["0","1"], types=[int,str])
    b.append([0,None])
    b.append([1,'b'])
    c = b[:]
    assert(tuple(b[0]) == tuple(c[0]))

    # Verify accessing multiple rows by list-access
    b = a[[0,2,3]]
    assert(tuple(b["0"]) == tuple([1,3,1]))

    # Verify load and save of a csv file
    a.save("a-test.csv")
    b = Data().load("a-test.csv")
    assert(tuple(a["0"]) == tuple(b["0"]))
    os.remove("a-test.csv")

    # Verify the writing of a CSV with quoted content and correct read.
    b = a[:]
    b[-1,1] = "string with comma, we'll see"
    b.save("a-test.csv")
    c = Data.load("a-test.csv")
    assert(tuple(c[-1]) == tuple(b[-1]))
    os.remove("a-test.csv")

    # Verify load and save of a pkl file
    a.save("a-test.pkl")
    b = Data.load("a-test.pkl")
    assert(tuple(a["0"]) == tuple(b["0"]))
    os.remove("a-test.pkl")

    # Verify load and save of a gzipped dill file
    try:
        a.save("a-test.dill.gz")
        b = Data().load("a-test.dill.gz")
        assert(tuple(a["0"]) == tuple(b["0"]))
        os.remove("a-test.dill.gz")
    except Data.MissingModule:
        print("(skipping dill test)..", end=" ")

    # Verify load and save of a gzipped csv file
    a.save("a-test.csv.gz")
    b = Data().load("a-test.csv.gz")
    assert(tuple(a["0"]) == tuple(b["0"]))
    os.remove("a-test.csv.gz")

    # Verify the behavior of the 'unpack' and 'pack' functions.
    b = a[:]
    b['iter'] = [None] + [list(range(i)) for i in range(len(b)-1)]
    _ = b.shape[1]
    b.unpack("iter")
    assert((b.shape[1] - _) == (len(b) - 3))
    b.pack("iter")
    assert(b.shape[1] == _)
    assert(b[0,-1] is None)
    assert(b[1,-1] is None)

    # TODO: Verify load of a large file (sample only)
    # TODO: Verify load of a large compressed file (sample only)

    # Verify column equivalence / equality checker by element
    b = a[a["0"] == 1]
    assert(tuple(b["0"]) == tuple([1,1]))

    # Verify one of the inequality operators
    b = a[a["0"] < 2]
    assert(tuple(b["0"]) == tuple([1,1,-1]))

    # Verify column membership checker by set
    b = a[a["1"] == {'a','b','2'}]
    assert(tuple(b["1"]) == tuple(['a','b','2']))

    # Verify column index can be an integer or a string.
    for i in range(len(a)):
        for j,name in enumerate(a.names):
            assert(a[i,j] == a[i,name])

    # Verify that the "in" operator works on rows.
    assert(1 in a[0])
    assert('a' in a[0])
    assert(1.2 in a[0])

    # Verify that "in" works on Data and View.
    assert(not ([1, "a"] in a))
    assert(not ([3, "a"] in a[:,:2]))
    assert([1, "a"] in a[:,:2])

    # Verify that "equality" works for rows.
    assert(a[0] == (1,'a',1.2))
    assert(not all(a[0] != a[0][:]))

    # Verify that "equality" works for Data objects.
    assert(tuple(a == [1,'a',1.2]) == (0,))

    # Verify that "equality" fails when the equality operator for
    # elements of a row is NotImplemented.
    b = a[:]
    assert( len(list(b == (i for i in (1,2,3,4)))) == 0 )
    assert( tuple(b[0] == (i for i in (1,2,3,4))) ==
            (NotImplemented, NotImplemented, NotImplemented) )

    # Verify that "addition" with sequences works correctly (left and right).
    b = a[:]
    b[:,1] = map(lambda v: (ord(v) if (v != None) else None), b[:,1])
    assert(sum(b[0] + [-1,-97,-1.2]) == 0)
    assert(sum([-1,-97,-1.2] + b[0]) == 0)
    assert(tuple(a[0] + a[0]) == (2,'aa',2.4))
    # Verify that "subtraction" with sequences works correctly (left and right).
    assert(sum(b[0] - [1,97,1.2]) == 0)
    assert(sum([1,97,1.2] - b[0]) == 0)

    # WARNING: Not individually verifying *all* comparison operators.

    # Verify generator-based index assignment.
    b = a[:]
    b[(i for i in range(0,len(b),2)), "1"] = "new"
    assert(b[0,"1"] == b[2,"1"] == b[4,"1"] == "new")

    # Verify list-based generator assignment.
    b = a[:]
    b[[i for i in range(0,len(b),2)], "1"] = str("test")
    assert(b[0,"1"] == b[2,"1"] == b[4,"1"] == "test")

    # Verify that assignment from a generator works.
    b = a[:]
    b['3'] = map(str, b['2'])
    assert(b[0,'3'] == str(b[0,'2']))
    assert(b[-1,'3'] == str(b[-1,'2']))

    # Verify assignment of a non-iterable being broadcast automatically.
    b['4'] = None
    assert(tuple(b[-2]) == (1,'2',3.0,'3.0',None))

    # Vefify addition of a column to empty data works.
    b = Data()
    b.add_column([1,2,3])
    assert(tuple(b[:,0]) == (1,2,3))

    # Verify the generation of a random k-fold cross validation.
    b = Data()
    for i in range(37):
        b.append([i])
    # Collect the list of all the "testing" rows seen
    all_test = []
    for (train, test) in b.k_fold(k=11, seed=1):
        all_test += list(test["0"])
    assert(sorted(all_test) == list(range(len(b))))

    # Verify the identification of unique rows and collection process.
    b = a[:]
    b += a
    cols = ['0','1']
    b = b[cols].unique().copy().collect(b)
    assert(tuple(b[0,-1]) == tuple(2*[a[0,-1]]))

    # Verify slicing data down to one column.
    assert(tuple(a[:,-1]) == (1.2,3.0,2.4,3.0,None))

    # Test slicing to access columns and rows.
    b = a[:,:-1]
    c = b.unique()
    assert(len(b) == len(c))

    # Test the access of a single column by name in a view.
    b = a[:,:-1]
    assert( tuple(b[:,"0"]) == (1,2,3,1,-1) )

    # Test assigning and updating a list typed column.
    b = Data(types=[str,list],data=[('a',['1']), ('b',['3'])])
    b[0,"1"] += [2.]
    b[1, 1 ] = b[1,1] + [4]
    assert(tuple(b[:,0]) == ('a', 'b'))
    assert(tuple(map(tuple,b[:,1])) == (('1',2.), ('3',4)))

    # Verify that the printed data set handles edge cases (one more
    # row or equal rows to number desired) correctly.
    b = a[:]
    b = b + b
    b.append(b[0])
    sb = str(b)
    # Test the correctness of truncated rows and columns.
    b.max_display = 2
    b_printout = '''
========================
Size: (11 x 3)          
                        
 0   | ... | 2          
 int | ... | float      
------------------      
 1   | ... | 1.2        
 ...   ...   ...        
 1   | ... | 1.2        
                        
 missing 4 of 33 entries
      at rows: [5, 10]  
   at columns: [1, 2]   
========================
'''
    assert( str(b) == b_printout)
    # Test the correctness of truncated rows only.
    b.max_display = 4
    b_printout = '''
========================
Size: (11 x 3)          
                        
 0   | 1    | 2         
 int | str  | float     
-------------------     
 1   | "a"  | 1.2       
 2   | "b"  | 3.0       
 ...   ...    ...       
 -1  | None | None      
 1   | "a"  | 1.2       
                        
 missing 4 of 33 entries
      at rows: [5, 10]  
   at columns: [1, 2]   
========================
'''
    assert( str(b) == b_printout)
    # Test the correctness of exact match (no truncation).
    b.max_display = 11
    b_printout = '''
========================
Size: (11 x 3)          
                        
 0   | 1    | 2         
 int | str  | float     
-------------------     
 1   | "a"  | 1.2       
 2   | "b"  | 3.0       
 3   | "c"  | 2.4       
 1   | "2"  | 3.0       
 -1  | None | None      
 1   | "a"  | 1.2       
 2   | "b"  | 3.0       
 3   | "c"  | 2.4       
 1   | "2"  | 3.0       
 -1  | None | None      
 1   | "a"  | 1.2       
                        
 missing 4 of 33 entries
      at rows: [5, 10]  
   at columns: [1, 2]   
========================
'''
    assert(str(b) == b_printout)

    # Verify that the names and types attributes are the correct type.
    assert("Descriptor" in str(type(a.types)))
    assert("Descriptor" in str(type(a.names)))

    # Make sure that two different methods of column access both return the same column.
    a1 = a[(i for i in range(len(a))), '1']
    a2 = a[(i for i in range(len(a)))]['1']
    assert(type(a1) == type(a2))
    assert(tuple(a1) == tuple(a2))

    # Check the 'flatten' function to make sure it correctly flattens
    # structures with nested, differential depth, and different types.
    complex_sequence = [(l for l in ((1,2,3)+tuple("abc"))),
                        "def", "ghi", ("jk",tuple("lmn"))]
    simplified_sequence = (1,2,3)+tuple('abcdefghijklmn')
    assert(tuple(flatten(complex_sequence)) == simplified_sequence)

    # Verify that merging two data sets works as expected.
    d1 = Data(
        [[0, 'a'], [1, 'b'], [2, 'b'], [7, 'd']],
        names = ['index', 'letter']
    )
    d2 = Data(
        [[None, 3, 'a'], [None, 4, 'b'], [None, 5, 'c'], [None, 6, 'b']],
        names = ['none', 'index', 'letter']
    )
    d3 = Data([
        ['a', 0, None, 3],
        ['b', 1, None, 4],
        ['b', 1, None, 6],
        ['b', 2, None, 4],
        ['b', 2, None, 6],
        ['c', None, None, 5],
        ['d', 7, None, None],
    ])
    out = merge_on_column(d1, d2, 'letter', d1_name="hoorah", d2_name="loopy")
    # Make sure the output columns are named as expected.
    assert(all(n1 == n2 for (n1, n2) in zip(
        out.columns, ['letter', 'index_hoorah', 'none', 'index_loopy'])))
    # Make sure all rows only show True or None in equality operator.
    assert(all(len(set(list(r1 == r2)) ^ {True,None}) == 0
               for (r1,r2) in zip(d3, out)))

    # Try initializing a data object with a generator (and having print statements).
    _ = Data(
        data=((j for j in range(10)) for i in range(100)),
    )

    # ----------------------------------------------------------------
    #     Testing expected exceptions.

    # Verify that indexing an element that doesn't exist fails.
    try: a.index([1,2,3])
    except ValueError: pass
    else: assert(False)

    # Verify that incorrect names are detected in "Data.effect".
    b = a[:]
    try: b.effect("bad name")
    except Data.BadIndex: pass
    else:  assert(False)

    # Verify that a data object refuses to be added to itself in place.
    b = a[:]
    try: b += b
    except Data.Unsupported: pass
    else:  assert(False)

    # Try assigning with a generator that is too short.
    try:   a['3'] = (i for i in range(len(a)-1))
    except Data.BadData: pass
    else:  assert(False)

    # Try assigning with a generator that is too long.
    try:   a['3'] = (i for i in range(len(a)+1))
    except Data.BadData: pass
    else:  assert(False)

    # Try providing a generator that does not give strictly integers
    try:   a[(v for v in ('a',1,'b'))]
    except Data.BadIndex: pass
    else: assert(False)

    # Try adding a too-short row
    b = a[:]
    try:   b.append([9,"z"])
    except Data.BadElement: pass
    else:  assert(False)

    # Try a too-short column
    b = a[:]
    try:   b.add_column([1])
    except Data.BadData: pass
    else:  assert(False)

    # Try a too-long column
    b = a[:]
    try:   b.add_column(map(float,range(1000)))
    except Data.BadData: pass
    else:  assert(False)

    # Try adding an empty column
    b = a[:]
    try:   b.add_column([])
    except Data.BadData: pass
    else:  assert(False)

    # Try adding a name that is not a string
    b = a[:]
    try:   b.add_column([1,2,3,4,5], name=1)
    except Data.BadSpecifiedName: pass
    else:  assert(False)

    # Try adding a column with multiple types
    b = a[:]
    try:   b.add_column([1,2,3,4,5.0])
    except Data.BadValue: pass
    else:  assert(False)

    # Try a mismatched-type slice assignment
    b = a[:]
    try:   b[:,0] = [1.0, 1, 1, 1, 1]
    except Data.BadAssignment: pass
    else:  assert(False)

    # Try a mismatched-type column assignment operation
    try:   a["0"] = list(a["0"])[:-1] + ["0"]
    except Data.BadAssignment: pass
    else:  assert(False)

    # Try a too-short column assignment operation
    try:   a["0"] = list(map(str,a["0"]))[:-1]
    except Data.BadValue: pass
    else: assert(False)

    # Try a too-long column assignment operation
    try:   a["0"] = list(map(str,a["0"])) + [1]
    except Data.BadAssignment: pass
    else:  assert(False)

    # Test for an error when an added data set has duplicate columns.
    b = a[:]
    c = a[:]
    c.names[2] = '1'
    try:    b.collect(c)
    except: Data.BadData
    else:   assert(False)

    # Try a bad combination of names and types
    try:   Data(names=["a","b","c"], types=[str, str])
    except Data.NameTypeMismatch: pass
    else:  assert(False)

    # Try a bad value in "names"
    try:   Data(names=[1,2,3], types=["a","b","c"])
    except Data.BadSpecifiedName: pass
    else:  assert(False)

    # Try a bad value in "types"
    try:   Data(names=["a","b","c"], types=[1,2,3])
    except Data.BadSpecifiedType: pass
    else:  assert(False)

    # Try bad value
    try:   a.append(["a","b","c"])
    except Data.BadValue: pass
    else:  assert(False)

    # Try bad element
    try:   a.append([1,2])
    except Data.BadElement: pass
    else:  assert(False)

    # Try non-existing column index
    try:   a["hello"]
    except Data.UnknownName: pass
    else:  assert(False)

    # Attempt to set item on an empty data
    b = Data()
    try:   b[0] = 1
    except Data.Empty: pass
    else:  assert(False)

    # Try get item on an uninitialized data
    try:   Data()["a"]
    except Data.Empty: pass
    else:  assert(False)

    # Try retyping an uninitialized data
    try:   Data().retype([])
    except Data.Empty: pass
    else:  assert(False)

    # Try reordering an uninitialized data
    try:   Data().reorder([])
    except Data.Empty: pass
    else:  assert(False)

    # Try popping from an uninitialized data
    try:   Data().pop()
    except Data.Empty: pass
    else:  assert(False)

    # Try saving an uninitialized data
    try:   Data().save("")
    except Data.Empty: pass
    else:  assert(False)

    # Done testing
    print("passed.")

    # Reset the maximum wait and the file argument for printing.
    Data.max_wait = _mw
    Data.print_kwargs.pop("file").close()
    if (_pf is not None):
        Data.print_kwargs["file"] = _pf



# =============================================================
#      END OF Python 'Data' with named and typed columns     
# =============================================================


# Run test cases
if __name__ == "__main__":
    # Run the tests without coverage.
    test_data()

    exit()

    # Try to run the tests with code coverage, otherwise just run them.
    try: 
        # Get the path to be watching for code coverage.
        import os
        path_to_watch = os.path.abspath(os.curdir)
        # Activate a code coverage module.
        import coverage
        cov = coverage.Coverage(source=[path_to_watch])
        cov.start()

        # Run the tests.
        test_data()

        # Save data from the coverage tracking (for report).
        cov.stop()
        cov.save()
        # Create a temporary directory for holding test results.
        from tempfile import TemporaryDirectory
        temp_dir = TemporaryDirectory()
        results_dir = temp_dir.name
        cov.html_report(directory=results_dir)
        # Open the results file.
        import webbrowser
        webbrowser.open("file://"+os.path.join(results_dir, "index.html"))
        # Wait for load, then delete the temporary directory.
        #   (user might decide to kill this process early).
        import time
        time.sleep(60*10)
        temp_dir.cleanup()
        del temp_dir
    except ModuleNotFoundError:
        pass



import os, random
import numpy as np
import ctypes
from tlux.approximate.axy.preprocessing import i_map, i_encode, to_array
from tlux.unique import unique, to_int
from tlux.profiling import profile



if __name__ == "__main__":
    random.seed()
    num_words = 10
    max_word_length = 16
    path = f"words-{max_word_length}_{num_words}.npz"
    if (not os.path.exists(path)):
        print(f"Generating {num_words} words..", flush=True)
        # Generate random words.
        words = [[
            "".join((chr(random.randint(ord('0'),ord('z')))
                     for _ in range(random.randint(0,max_word_length))))
              .replace("\\", "/")
              .replace("`", ".")
            , random.randint(0, max_word_length**2)] for _ in range(num_words)
        ]
        # Convert those words to a numpy array.
        xi = np.asarray(words, dtype=object)
        # Save the words to file (for later reuse).
        np.savez_compressed(path, xi=xi)
    else:
        print(f"Loading '{path}'..", flush=True)
        xi = np.load(path, allow_pickle=True)['xi']

    print("Modifying..", flush=True)
    xi = xi.reshape((-1,2))

    # @profile
    # def test(steps=3):
    #     for _ in range(steps):
    #         print("", _, end="\r")
    #         res = unique(xi)
    #         del(res)
    # for _ in range(10):
    #     test()
    #     test.show_profile()
    # print()
    # unique.show_profile()
    # exit()

    # Convert data into the xi_map.
    print("Calling i_map..", flush=True)
    xi_map = i_map(xi, [])
    if (hasattr(unique, "show_profile")): unique.show_profile()
    if (hasattr(i_map, "show_profile")): i_map.show_profile()

    for i,m in enumerate(xi_map):
        print(f" map {i} (len {len(m)})", m)

    # Encode the data in integer format.
    print("Calling i_encode..", flush=True)
    final_xi = i_encode(xi, xi_map)
    if (hasattr(to_int, "show_profile")): to_int.show_profile()
    if (hasattr(i_encode, "show_profile")): i_encode.show_profile()

    # Print the source data and the final integer version.
    print(xi)
    print(final_xi)



'''--------------------------------------------------------------------
                            regex.py
 
 A fast regular expression matching code for Python, built on top of
 the 'regex.c' library. The main function provided here is:

   match(regex, string) -> (start, end) or None or RegexError,

 Documentation for 'regex.c' follows.

''' 


#  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
#                         Python development
# 
# 
#   write `fmatch` function to go with the existing
#   `match`, `matcha`, and `fmatcha` functions.
# 
#   add parallelism into the C library code, including a file walk,
#   to offer the full `frex` stack within the C code.
# 
#   fix all documentation to align with the current structure
# 
#   write a C function for translating from classic regular
#   expressions into fast regular expressions
# 
#   default regex library has a `finditer` command that iterates over
#   matches and streams bytes to the regular experssion library from
#   a buffer, can that be done here?
#   ^^ write a smatch and sfmatch that do 'stream' style matching,
#      being given a start index and number of bytes to parse,
#      maintaining a global state that is used for repeated calls to
#      the same C function
# 
#   create a Python test case that would reveal any memory leak in the
#   returned allocatable arrays from the C library
# 
#   look into building c extensions automatically on any platform,
#   some compilation customization will need to be done for windows
#   ^^ should I support something I don't plan on using myself?
# 
#   write a "summary" of the algorithm in HTML with embedded SVG files
#   and a theoretic description
# 
#   build some other code that uses "match" and make sure it works as
#   would be expected, measure performance to see what's slow
# 
#   use Pool.map instead of the homegrown solution, remove dependency
#   on the parallel map code <- the Pool.map solution is slower, so
#   maybe it's best to use the custom matching code. Otherwise, it
#   might be best to just use `os.fork` and manually create all of
#   the multiprocessing environment. Then this code will definitely
#   not work on Windows.
# 
#   convert things that are not bytes nor strings into strings using
#   the "str" function inside of `match`?
#
#   make the test code automatically run when the code is compiled,
#   checking for correct compilation
# 
#   regex line based mode from the command line
# 
#   regex invert match from the command line (default to activate line based mode)
# 
#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# 
#                            C development
# 
#   The token following a ".*" can get excluded from the final match,
#    because the 'start index' of that token is updated before the
#    token after it is matched and the regex is completely discovered.
#    Use a queue to store the tokens instead of stack (track start and
#    end instead of just count. Special tokens should be given top
#    priority, then character sets.
# 
#   If the "start" index hasn't changed, don't add another match (for
#   `matcha` type operations, causes too many overlapping matches).
# 
#  Make this a pure C executable, place all Python functionality into
#   this file when compiled (without DEBUG).
# 
#  Swap out DEBUG for DEVELOP.
# 
#  Refactor into functions:
#    match -- find first regex match in a string
#    matcha -- find all regex matches in a string
#    matchs -- stream string matches by making repeated calls to same function
#    fmatch -- find the first regex match in a file
#    fmatcha -- find all regex matches in a file
#    fmatchs -- stream file matches by making repeated calls to same function
#    imatch -- find the first match from standard input and exit
#    imatcha -- find all matches from standard input
#    imatchs -- stream matches from standard input
#    frex -- file regular experession search utility, like `grep` but better
#            parallel (num cores) or serial
#            recurse into directories or not
#            provide path names or path patterns
#            optional redirect from standard in input
#            translate (most) standard regexes or give error message
#            parallel `fmatchs` (`imatchs`) that prints output to
#              standard output, or specified file object
# 
#  Write tests for `fmatch`.
# 
#  Rescan comments and update all documenation throughout.
# 
#  Identify ways to reduce size of code within reason, should some
#   things be abstracted as functions to save repetition?
# 
#  Write `randre` that generates a random string that matches the
#   given regular expression, tile the space of matches so that
#   all the shortest matches come first, then longer ones
#  
#  Find way to pass in arbitrary "next token" function to the match
#   function, that way `match` and `fmatch` can all go through one
#   function whether given a character array or a file
# 
#  ___________________________________________________________________


PRINT_PREFIX = "# "
PRINT_FILE_SEPARATORS = False
HALF_MAX_PREVIEW_WIDTH = 20

# Import ctypes for loading the underlying C regex library.
import os, ctypes

# --------------------------------------------------------------------
#                 Darwin (macOS) / Linux (Ubuntu) import
_clib_bin = os.path.join(os.path.dirname(os.path.abspath(__file__)), "libregex.so")
# Import or compile the C file.
try:
    clib = ctypes.CDLL(_clib_bin)
except:
    from tlux.setup import build_regex
    clib = build_regex()
# --------------------------------------------------------------------


# Exception to raise when errors are reported by the regex library.
class RegexError(Exception): pass


# Given a regular expression in a Unix-like format, translate it to a
# regular experssion that is (roughly) equivalent in the language of
# the `regex.c` library.
def translate_regex(regex, case_sensitive=True):
    # Do substitutions that make the underlying regex implementation
    # behave more like common existing regex packages.
    if (len(regex) > 0):
        # Add a ".*" to the front of the regex if the beginning of the
        # string was not explicitly desired in the match pattern.
        if (regex[0] == "^"): regex = regex[1:]
        elif ((len(regex) < 2) or (regex[0] != ".") or (regex[1] != '*')): 
            regex = ".*" + regex
        # Add a "{.}" to the end of the regex if the end of the string
        # was explicitly requested in the pattern.
        if (regex[-1] == "$"): regex = regex[:-1] + "{.}"
    # Replace all alphebetical characters with token sets that include
    # all cases of that character.
    if (not case_sensitive):
        i = 0
        in_literal = False
        while (i < len(regex)):
            # Check for the beginning of a token set.
            if ((regex[i] == '[') and (not in_literal)):
                in_literal = True
                contains = set()
                missing = set()
            # Handle a currently-active token set.
            elif (in_literal):
                # Check for the end of this token set.
                if (regex[i] == ']'):
                    in_literal = False
                    # Add all the missing characters to this token set.
                    missing = ''.join(sorted(missing.difference(contains)))
                    regex = regex[:i] + missing + regex[i:]
                    # Increment i appropriately.
                    i += len(missing)-1
                # Add the paired-case if we see a cased character.
                elif (regex[i].isalpha()):
                    contains.add(regex[i])
                    if (regex[i].islower()):
                        missing.add(regex[i].upper())
                    else:
                        missing.add(regex[i].lower())
            # Otherwise replace this single character with a token set.
            elif (regex[i].isalpha()):
                if (regex[i].islower()):
                    token_set = f'[{regex[i]}{regex[i].upper()}]'
                else:
                    token_set = f'[{regex[i]}{regex[i].lower()}]'
                regex = regex[:i] + token_set + regex[i+1:]
                i += len(token_set)-1
            # Increment to the next token.
            i += 1
    # Return the now-prepared regular expression.
    return regex


# Translate 'start' and 'end' values that are returned by the `regex.c`
# library, raising appropriate errors as defined by the library.
def _translate_return_values(regex, start, end):
    # Return appropriately (handling error flags).
    if (start >= 0): return (start, end)
    elif (start < 0):
        if (end == 0): return None # no match found
        elif (end == -1): return (0, 0) # empty regular expression
        elif ((start == -1) and (end == -5)): return None  # empty string
        elif (end < -1): # error code provided by C
            err = f"Invalid regular expression (code {-end})"
            if (start < -1):
                start -= sum(1 for c in str(regex,"utf-8") if c in "\n\t\r\0")
                err += f", error at position {-start-1}.\n"
                err += f"  {str([regex])[1:-1]}\n"
                err += f"  {(-start)*' '}^"
            else: err += "."
            raise(RegexError(err))


# Find a match for the given regex in string.
# 
#   match(regex, string) -> (start, end) or None or RegexError,
# 
# where "regex" is a string defining a regular expression and "string"
# is a string to be searched against. The function will return None
# if there is no match found. If a match is found, a tuple with
# integers "start" (inclusive index) and "end" (exclusive index) will
# be returned. If there is a problem with the regular expression,
# a RegexError will be raised.
#
# Some substitutions are made before passing the regular expressions
# "regex" to the "match" function in 'regex.c'. These substitutions
# include:
#
#  - If "^" is at the beginning of "regex", it will be removed (because
#    the underlying library implicitly assumes beginning-of-string.
#  - If no "^" is placed at the beginning of "regex", then ".*" will
#    be appened to the beginning of "regex" to behave like other 
#    regular expression libraries.
#  - If "$" is the last character of "regex", it will be substituted
#    with "{.}", the appropriate pattern for end-of-string matches.
# 
def match(regex, string, **translate_kwargs):
    # Translate the regular expression to expected syntax.
    regex = translate_regex(regex, **translate_kwargs)
    # Call the C utillity.
    #   initialize memory storage for the start and end of a match
    start = ctypes.c_int()
    end = ctypes.c_int()
    #   convert strings into character arrays
    if (type(regex) == str): regex = regex.encode("utf-8")
    if (type(string) == str): string = string.encode("utf-8")
    c_regex = ctypes.c_char_p(regex);
    c_string = ctypes.c_char_p(string);
    #   execute the C function
    clib.match(c_regex, c_string, ctypes.byref(start), ctypes.byref(end))
    del(c_regex, c_string, string)
    # Return the values from the C library (translating them appropriately)
    return _translate_return_values(regex, start.value, end.value)

# Get all matches for a regex.
def matcha(regex, string, **translate_kwargs):
    # Use a C utility to scal the file for all matches.
    import ctypes
    # Translate the regular expression to expected syntax.
    regex = translate_regex(regex, **translate_kwargs)
    # Call the C utillity.
    #   initialize memory storage for the start and end of a match
    n = ctypes.c_int()
    starts = ctypes.c_void_p()
    ends = ctypes.c_void_p()
    #   convert strings into character arrays
    if (type(regex) == str): regex = regex.encode("utf-8")
    if (type(string) == str): string = string.encode("utf-8")
    c_regex = ctypes.c_char_p(regex);
    c_string = ctypes.c_char_p(string);
    #   execute the C function
    clib.matcha(c_regex, c_string, ctypes.byref(n),
                ctypes.byref(starts), ctypes.byref(ends))
    # Get the number of matches.
    n = n.value
    # Return the values from the C library (translating them appropriately)
    if (n == 0): return [], []
    elif (n < 0):
        if (n == -2): raise(TypeError("`matcha` must be provided with a nonempty string."))
        elif (n == -1):
            starts = (ctypes.c_int*1).from_address(starts.value)
            ends = (ctypes.c_int*1).from_address(ends.value)
            result = _translate_return_values(regex, starts[0], ends[0])
            # ^^ this might raise an error, otherwise it's a no match
            if (result == None):    return [], []
            elif (result == (0,0)): raise(RegexError("`matcha` requires nonempty regular expression."))
            else:                   raise(NotImplementedError)
    else:
        starts = (ctypes.c_int*n).from_address(starts.value)
        ends = (ctypes.c_int*n).from_address(ends.value)
    return list(starts), list(ends)


# Given a path to a file, search for all nonoverlapping matches in the
# file and return the count of number of matches and a summary string.
def fmatcha(path, regex, ascii_ratio=0.7, **translate_kwargs):
    # Make sure the file exists.
    if (not os.path.exists(path)): return path, 0, ""
    # Use a C utility to scal the file for all matches.
    import ctypes
    # Translate the regular expression to expected syntax.
    regex = translate_regex(regex, **translate_kwargs)
    # Call the C utillity.
    #   initialize memory storage for the start and end of a match
    n = ctypes.c_int()
    starts = ctypes.c_void_p()
    ends = ctypes.c_void_p()
    lines = ctypes.c_void_p()
    min_ascii_ratio = ctypes.c_float(ascii_ratio)
    #   convert strings into character arrays
    if (type(regex) == str): regex = regex.encode("utf-8")
    if (type(path) == str): path = path.encode("utf-8")
    c_regex = ctypes.c_char_p(regex);
    c_path = ctypes.c_char_p(path);
    #   execute the C function
    clib.fmatcha(c_regex, c_path, ctypes.byref(n),
                 ctypes.byref(starts), ctypes.byref(ends),
                 ctypes.byref(lines), min_ascii_ratio)
    n = n.value
    path = str(path, 'utf-8')
    # Return the values from the C library (translating them appropriately)
    if (n == 0):
        return path, 0, f"  no matches in '{path}'"
    elif (n < 0):
        if (n == -3):
            return path, 0, f"  binary file skipped at '{path}'"
        if (n == -2):
            raise(OSError(f"Failed to open file '{path}'."))
        elif (n == -1):
            starts = (ctypes.c_int*1).from_address(starts.value)[0]
            ends = (ctypes.c_int*1).from_address(ends.value)[0]
            result = _translate_return_values(regex, starts, ends)
            # ^^ this might raise an error, otherwise it's a no match
            if (result == None): return path, 0, f"  no matches in '{path}'"
            elif (result == (0,0)):
                raise(RegexError("`fmatcha` requires nonempty regular expression."))
            else:
                raise(NotImplementedError)
    else:
        # cd ~/Git/Old/VarSys/ ; frex "poetry"
        # cd ~/Git/regex ; frex -s "he" "[.]py"

        if PRINT_FILE_SEPARATORS: summary = f"{'-'*70}\n{' '*(32-len(path)//2)}{path}"
        else: summary = ""
        starts = (ctypes.c_int*n).from_address(starts.value)
        ends = (ctypes.c_int*n).from_address(ends.value)
        lines = (ctypes.c_int*n).from_address(lines.value)
        with open(path, "rb") as f:
            for i in range(n):
                f.seek(max(0,starts[i]-HALF_MAX_PREVIEW_WIDTH), 0)
                bytes_to_read = (ends[i] - starts[i]) + 2*HALF_MAX_PREVIEW_WIDTH
                if (bytes_to_read < 1):
                    print("ERROR:")
                    print("list(starts): ",list(starts))
                    print("list(ends):   ",list(ends))
                    print("list(lines):  ",list(lines))
                    print("i: ",i)
                    print("bytes_to_read: ",bytes_to_read)
                    raise(Exception())
                match_line_string = str(f.read(bytes_to_read))[1:]
                if PRINT_FILE_SEPARATORS: summary += f"\n{lines[i]}: {match_line_string}"
                else: summary += f"\n{PRINT_PREFIX}{path}:{lines[i]:<5d} {match_line_string}"
        return path, n, summary


# Do a fast regular expression search over files that match a given
# pattern. Find all nonoverlapping matches in the files and print
# all matching patterns, their files, and their locations.
def frex(regex, *path_patterns, curdir=".", recursive=True, 
         parallel=True, chunksize=10, **translate_kwargs):
    # Get all candidate paths that *might* be searched.
    if recursive:
        candidate_paths = os.walk(curdir)
    else:
        candidate_paths = ((curdir,None,[p]) for p in os.listdir(curdir)
                           if os.path.isfile(os.path.join(curdir,p)))
    # Determine searchable paths from candidate paths.
    paths = []
    for (dirname, _, file_paths) in candidate_paths:
        for fname in file_paths:
            path = os.path.join(dirname, fname)
            for pattern in path_patterns:
                if (match(pattern, path, **translate_kwargs) is not None): break
            else: continue
            paths.append(path)
    # Perform the search over all the candidate paths (in parallel).
    if parallel:
        from functools import partial
        from multiprocessing import Pool, cpu_count
        pool = Pool(cpu_count())
        translate_kwargs["regex"] = regex
        p_fmatcha = partial(fmatcha, **translate_kwargs)
        match_iterator = pool.imap_unordered(p_fmatcha, paths, chunksize=chunksize)
    else:
        match_iterator = (fmatcha(p, regex, **translate_kwargs) for p in paths)
    # Cycle over all matches and print the summaries.
    matches = {}
    for p,n,s in match_iterator:
        matches[p] = n
        if (n > 0): print(s)
    # Clean up parallel Pool object.
    if parallel:
        pool.close()
        pool.terminate()
    # Return the dictionary containing all paths and matches.
    return matches


# Main for when this is executed as a program.
def main():
    import sys
    # Extract the "not recursive" optional flag if it exists.
    if ("-n" in sys.argv):
        recursive = False
        sys.argv.remove("-n")
    else: recursive = True
    # Extract the "case sensitive" optional flag if it exists.
    if ("-c" in sys.argv):
        case_sensitive = True
        sys.argv.remove("-c")
    else: case_sensitive = False
    # Extract the "serial" optional flag if it exists.
    if ("-s" in sys.argv):
        serial = True
        sys.argv.remove("-s")
    else: serial = False
    # Check for proper usage.
    if (len(sys.argv) < 2):
        print(f'''
ERROR: Only {len(sys.argv)} command line argument{'s' if len(sys.argv) > 1 else ''} provided.

Expected call to look like:
  python3 -m regex [-n] [-c] [-s] "<search-pattern>" ["<path-pattern-1>"] ["<path-pattern-2>"] [...]

"-n" is provided if the call to `frex` should NOT recursively
search all files in the directory tree from the current directory.
If no path patterns are given, all files are searched.
Documentation for this module follows.

"-c" is provided if the given regular expression is case sensitive,
meaning the letters should be matched with the exact same case.

"-s" is provided if the search should be run serially (not in parallel).

See `python -c "import regex; help(regex)"` for more detailed 
documentation including the regular expression language specification.
''')
        exit()
    regex = sys.argv[1]
    print("Given regex:",str([regex])[1:-1])
    print("Using regex:",str([translate_regex(regex,case_sensitive)])[1:-1])
    path_patterns = sys.argv[2:]
    # Set the default path pattern to match all paths.
    if (len(path_patterns) == 0): path_patterns = [""]
    curdir = os.path.abspath(os.path.curdir)
    # Do a fast regular expression search.
    matches = frex(regex, *path_patterns, curdir, recursive=recursive,
                   case_sensitive=case_sensitive, parallel=(not serial))
    total_matches = sum(matches.values())
    if (total_matches > 0):
        print(f"\n found {total_matches} match{'es' if total_matches > 0 else ''} across {len(matches)} files")
    else:
        print(f"\n no matches found across {len(matches)} files")


# Append the documentation from "regex.c" to the documentation for this module.
with open(os.path.join(os.path.dirname(__file__), "regex.c")) as _f:
    _source_file = _f.read()
    _start, _end = match("^/.*\n{/}", _source_file)
    __doc__ += _source_file[_start+3:_end].replace("\n//", "\n")
    del(_source_file, _start, _end)
del(_f)


# When using "from regex import *", only get these variables:
__all__ = ["RegexError", "match", "frex", "match", "matcha", "main"]


if __name__ == "__main__":
    import cProfile as profile
    profile.run("main()", sort='time') # sort by total time
    # profile.run("main()", sort='cumtime') # sort by cumulative time



import os
from tlux.approximate.axy import AXY
import numpy as np

# Import codes that will be used for testing.
from tlux.plot import Plot
from tlux.random import well_spaced_box

seed = 0
np.random.seed(seed)

n = 2**8
d = 2
new_model = True
use_a = True
agg_dim = 64
agg_states = 4
use_x = False
model_dim = 64
model_states = 4
model_dim_output = 0
use_yi = False
steps = 1000
num_threads = None

ONLY_SGD = dict(
    faster_rate = 1.0,
    slower_rate = 1.0,
    update_ratio_step = 0.0,
    step_factor = 0.001,
    step_mean_change = 0.1,
    step_curv_change = 0.01,
    keep_best = True,
    basis_replacement = False,
)

settings = dict(
    seed=seed,
    early_stop = False,
    logging_step_frequency = 1,
    rank_check_frequency = 100,
    **({"mdo":model_dim_output} if model_dim_output is not None else {}),
    normalize_x = True,
    normalize_y = True,
    # **ONLY_SGD
)

# A function for testing approximation algorithms.
def f1(x):
    x = x.reshape((-1,2))
    x, y = x[:,0], x[:,1]
    return (3*x + np.cos(8*x)/2 + np.sin(5*y))

# A second function for testing approximation algorithms.
def f2(x):
    x = x.reshape((-1,2))
    return np.cos(np.linalg.norm(x,axis=1))

# A third function that is the trigonometric shift of the first.
def f3(x):
    x = x.reshape((-1,2))
    x, y = x[:,0], x[:,1]
    return (3*x + np.sin(8*x)/2 + np.cos(5*y))


functions = [f1, f2] # , f3]

# Generate data bounds.
x_min_max = [[-.1, 1.1], [-.1, 1.1]]

# Create all data.
points = [well_spaced_box(n, d) for _ in functions]
x =  np.concatenate(points, axis=0).astype("float32")
xi = np.asarray([n*[f.__name__] for f in functions], dtype=object).reshape((-1,1))
ax = x.reshape((-1,1)).copy()
axi = np.asarray(list(zip(
    [f.__name__ for f in functions for _ in range(n*d)],
    [i+1 for _ in range(n*len(functions)) for i in range(d)]
)), dtype=object)
sizes = np.ones(n*len(functions), dtype="int32") * d

# Concatenate the two different function outputs.
y = np.concatenate([f(p) for (f,p) in zip(functions, points)], axis=0).reshape((len(functions)*n,1))
# Generate classification data that is constructed by binning the existing y values.
yi = np.asarray([
    np.where(
        y[:,0] <= np.percentile(y[:,0], 50),
        'bottom',
        'top'
    ),
    np.where(
        y[:,0] <= np.percentile(y[:,0], 20),
        'small',
        np.where(
            y[:,0] <= np.percentile(y[:,0], 80),
            'medium',
            'large'
        )
    ),
], dtype=object).T
# Compute the numeric values associated with each Y group (mean of values in group).
yi_values = [{
    "bottom": y[y[:,0] <= np.percentile(y[:,0], 50)].mean(),
    "top": y[y[:,0] > np.percentile(y[:,0], 50)].mean(),
}, {
    "small": y[y[:,0] <= np.percentile(y[:,0], 20)].mean(),
    "medium": y[(y[:,0] > np.percentile(y[:,0], 20)) *
                (y[:,0] < np.percentile(y[:,0], 80))].mean(),
    "large": y[y[:,0] >= np.percentile(y[:,0], 80)].mean(),
}]

# Train a new model or load an existing one.
if (new_model or (not os.path.exists('temp-model.json'))):
    # Initialize a new model.
    print("Fitting model..")
    m = AXY(
        ads=agg_dim,
        ans=agg_states,
        mds=model_dim,
        mns=model_states,
        num_threads=num_threads,
        **settings,
    )
    m.fit(
        ax=(ax.copy() if use_a else None),
        axi=(axi if use_a else None),
        sizes=(sizes if use_a else None),
        x=(x.copy() if use_x else None),
        xi=(xi if use_x else None),
        y=y.copy(),
        yi=(yi if use_yi else None),
        steps=steps,
    )
    # Save and load the model.
    m.save("temp-model.json")

# Load the saved model.
m = AXY()
m.load("temp-model.json")
# Remove the saved model if only new models are desired.
if new_model:
    os.remove("temp-model.json")

# Print the model.
print()
print(m, m.config)
print()

def obj(params):
    m.model[:] = np.asarray(params, dtype="float32")
    # Evaluate the model at all training data.
    _ax = ax.copy()
    _axi = axi
    _sizes = sizes
    _x = x.copy()
    _xi = xi
    if (not use_a): _ax = _axi = _sizes = None
    if (not use_x): _x = _xi = None
    output = m.predict(ax=_ax, axi=_axi, sizes=_sizes, x=_x, xi=_xi)
    error = ((y - output)**2).sum()
    return float(error)

# --------------------------------------------------------
# Training by gradient free optimization.
# 
# The better this approach does, the worse the original trianing for the model performed.
# 
from util.optimize import minimize
sol = minimize(
    obj,
    solution=m.model.tolist(),
    bounds=[(-10,10)]*len(m.model),
    display=True,
    max_time=5,
)
m.model[:] = sol[:]
# --------------------------------------------------------

# Define a function that evaluates the model with some different presets.
def fhat(x, f=functions[0].__name__, yii=None):
    n = x.shape[0]
    xi = np.asarray([f]*n, dtype=object).reshape((-1,1))
    ax = x.reshape((-1,1)).copy()
    axi = np.asarray(list(zip(
        [f for _ in range(n*d)],
        [i+1 for _ in range(n) for i in range(d)]
    )), dtype=object)
    sizes = np.ones(x.shape[0], dtype="int32") * d
    if (not use_a): ax = axi = sizes = None
    if (not use_x): x = xi = None
    if (not use_yi): yii = None
    if (yii is None): return m(ax=ax, axi=axi, sizes=sizes, x=x, xi=xi)[:,0]
    else:             return [yi_values[yii][v] for v in m(ax=ax, axi=axi, sizes=sizes, x=x, xi=xi)[:,y.shape[1]+yii]]

# Generate the plot of the results.
print("Adding to plot..")
p = Plot()

# Add the provided points.
for i, (f, xf) in enumerate(zip(functions, points)):
    p.add(f"xi={f.__name__}", *xf.T, f(xf), color=i, group=i)

# Add the two functions that are being approximated.
for i,f in enumerate(functions):
    f = f.__name__
    p.add_func(f"xi={f}", lambda x: fhat(x, f=f), *x_min_max,
               group=i,  vectorized=True, color=i, opacity=0.8,
               plot_points=3000) # , mode="markers", marker_size=2)
if (use_yi):
    for f in functions:
        for i in range(yi.shape[1]):
            p.add_func(f"xi={f.__name__}, yi={i}", lambda x: fhat(x, f=f.__name__, yii=i),
                       *x_min_max, opacity=0.8, plot_points=3000,
                       mode="markers", shade=True, marker_size=4)

# Produce the visual.
print("Generating surface plot..")
# Generate a visual of the loss function.
if (len(getattr(globals().get("m",None), "record", [])) > 0):
    p.plot(show=False)
    print()
    print("Generating loss plot..")
    p = Plot("Mean squared error")
    # Rescale the columns of the record for visualization.
    record = m.record
    for i in range(0, record.shape[0]+1, max(1,record.shape[0] // 100)):
        step_indices = list(range(1,i+1))
        p.add("MSE", step_indices, record[:i,0], color=1, mode="lines", frame=i)
        p.add("Step factors", step_indices, record[:i,1], color=2, mode="lines", frame=i)
        p.add("Step sizes", step_indices, record[:i,2], color=3, mode="lines", frame=i)
        p.add("Update ratio", step_indices, record[:i,3], color=4, mode="lines", frame=i)
        p.add("Eval utilization", step_indices, record[:i,4], color=5, mode="lines", frame=i)
        p.add("Grad utilization", step_indices, record[:i,5], color=6, mode="lines", frame=i)
    p.show(append=True, show=True, y_range=[-.2, 1.2])
else:
    p.show()
print("", "done.", flush=True)



#!python 

# Any python code that should be executed (as if __main__) during
# setup should go here to prepare the module on a new computer.


# Function to run during module setup.
def setup():
    # Try compiling the 'fmath' library.
    import sys
    _first = sys.path.pop(0)
    try:
        build_axy()
        build_balltree()
        build_delaunay()
        build_regex()
        build_unique()
    except Exception as exc:
        print(f"WARNING: 'tlux' encountered exception while trying to build compiled libraries.")
        print()
        print(exc)
    finally:
        sys.path.insert(0, _first)


# Build the AXY model (compiled Fortran code).
def build_axy():
    import os, fmodpy
    # Define the output directory.
    _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "axy")
    # Import the RANDOM module, it is self contained and simple.
    _random = fmodpy.fimport(
        input_fortran_file = os.path.join(_dir, "axy_random.f90"),
        dependencies = ["pcg32.f90"],
        output_dir = _dir
    )
    # Import the AXY module, it has many dependencies and uses BLAS and LAPACK libraries.
    #   Get the directory for the AXY compiled source code.
    _dependencies = ["pcg32.f90", "axy_random.f90", "axy_matrix_operations.f90", "axy_sort_and_select.f90", "axy.f90"]
    _path = os.path.join(_dir, "axy.f90")
    _axy_kwargs = dict(
        input_fortran_file = _path,
        dependencies = _dependencies,
        output_dir = _dir,
        blas = True,
        lapack = True,
        omp = True,
        wrap = True,
        verbose = False,
    )
    #   Try and build using (local) defaults for fmodpy compilation.
    try:
        _axy = fmodpy.fimport(**_axy_kwargs)
        # Try calling the underlying library.
        _axy.axy.new_model_config(
            adn=2, ado=3, ads=4, ans=1, ane=6, ade=7,
            mdn=8, mdo=9, mds=10, mns=1, mne=12, mde=13,
            num_threads=1
        )
    #   When local default fail, try specifying exactly what shared library functions are needed.
    except Exception as exc:
        import logging
        logging.warn(f"Encountered execption with default build of AXY, trying again with an explicit library search.\n  {exc}")
        _axy = fmodpy.fimport( **_axy_kwargs,
            link_blas="", link_lapack="", rebuild=True,
            libraries = [_dir] + fmodpy.config.libraries,
            symbols = [
                ("sgemm", "blas"),
                ("sgels", "lapack"),
                ("omp_get_max_threads", "omp")
            ],
        )

    # Return the AXY and RANDOM modules.
    return _axy, _random


# Build the BallTree code.
def build_balltree():
    # from util.approximate.delaunay import delsparse
    import os, fmodpy
    _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "balltree")
    # Ball tree.
    _dependencies = ["swap.f90", "prune.f90", "fast_select.f90", "fast_sort.f90", "ball_tree.f90"]
    _path = os.path.join(_dir, "ball_tree.f90")
    _balltree = fmodpy.fimport(
        _path, dependencies=_dependencies, output_dir=_dir, omp=True,
        libraries = [_dir] + fmodpy.config.libraries,
        symbols = [("omp_get_max_threads", "omp")],
    )
    # Fast sort.
    _dependencies = ["swap.f90", "fast_sort.f90"]
    _path = os.path.join(_dir, "fast_sort.f90")
    _fast_sort = fmodpy.fimport(_path, dependencies=_dependencies, output_dir=_dir)
    # Fast select.
    _dependencies = ["swap.f90", "fast_select.f90"]
    _path = os.path.join(_dir, "fast_select.f90")
    _fast_select = fmodpy.fimport(_path, dependencies=_dependencies, output_dir=_dir)
    # Prune.
    _path = os.path.join(_dir, "prune.f90")
    _prune = fmodpy.fimport(_path, output_dir=_dir)
    return _balltree.ball_tree, _fast_sort.fast_sort, _fast_select.fast_select, _prune.prune


# Build the Delaunay code.
def build_delaunay():
    # from util.approximate.delaunay import delsparse
    import os, fmodpy
    _dependencies = ["blas.f", "lapack.f", "slatec.f", "delsparse.f90"]
    _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "delaunay")
    _path = os.path.join(_dir, "delsparse.f90")
    _delsparse = fmodpy.fimport(
        _path, dependencies=_dependencies, output_dir=_dir,
        omp=True, f_compiler_args="-std=legacy -fPIC -shared -O3",
    )
    return _delsparse
    

# Build the regex library.
def build_regex():
    import os, ctypes
    # Configure for the compilation for the C code.
    _c_compiler = "cc"
    _clib_source = os.path.join(os.path.dirname(os.path.abspath(__file__)), "regex", "regex.c")
    _clib_bin = os.path.join(os.path.dirname(os.path.abspath(__file__)), "regex", "libregex.so")
    try:
        # Import existing compiled library.
        _regex = ctypes.CDLL(_clib_bin)
    except:
        # Compile and import.
        _compile_command = f"{_c_compiler} -O3 -fPIC -shared -o '{_clib_bin}' '{_clib_source}'"
        os.system(_compile_command)
        _regex = ctypes.CDLL(_clib_bin)
    # Return the compiled module.
    return _regex


# Build the unique library.
def build_unique():
    import os, ctypes, fmodpy
    # Configure the compilation.
    _verbose = True
    # _c_compiler = "cc"  # <- on macOS with Clang, no nested functions allowed.
    _c_compiler = fmodpy.config.f_compiler  # <- works with gfortran, assuming true for others too.
    _flags = "-shared -fPIC -O3"  # <- shared library, position independent, O3 optimization.
    _flags += " -fopenmp"  # <- slower for a lot of cases, only good for *very* large data.
    # _flags += " -D_UNIQUE_SINGLE_THREADED"  # <- disables OpenMP parallelism and imports.
    _flags += " -D_UNIQUE_ALLOW_LOCAL_FUNCTIONS"  # <- enable if compiler supports local functions for variable width comparison.
    # _flags += " -D_UNIQUE_DEBUG_ENABLED"  # <- enable for testing / debugging.
    _lib_dir = os.path.dirname(os.path.abspath(__file__))
    _lib_path = os.path.join(_lib_dir, "unique", "unique.so")
    _src_path = os.path.join(_lib_dir, "unique", "unique.c")
    try:
        _unique = ctypes.cdll.LoadLibrary(_lib_path)
    except:
        # Compile the shared library.
        import sys
        cmd = f'{_c_compiler} {_flags} -o "{_lib_path}" "{_src_path}"'
        if _verbose: print(" ", cmd, flush=True)
        os.system(cmd)
        if _verbose: print(flush=True)
        # Load the shared library
        _unique = ctypes.cdll.LoadLibrary(_lib_path)
    # Return the compiled module.
    return _unique



from tlux.unique import unique, to_int
import multiprocessing
import numpy as np


# Convert a numpy array to a set of unique values.
def to_set(array): return set(array.tolist())


# Insert 'value' into 'array' in sorted order, left of anything
#  with equal 'key' if 'unique=False', otherwise skipped for equal keys.
def insert_sorted(array, value, key=lambda i:i, unique=True):
    low = 0
    high = len(array)
    index = (low + high) // 2
    while high != low:
        if key(array[index]) > key(value):   high = index
        elif key(array[index]) < key(value):  low = index + 1
        elif unique: return
        index = (low + high) // 2
    if ((len(array) == 0) or (array[index] != value) or (not unique)):
        array.insert(index, value)


# Given a categorical input array, construct a dictionary for
#  mapping the unique values in the columns of the array to integers.
def i_map(xi, xi_map):
    # Generate the map (ordered list of unique values).
    if (len(xi.dtype) > 0):
        for n in xi.dtype.names:
            xi_map.append(xi[n]) 
    else:
        for i in range(xi.shape[1]):
            xi_map.append(xi[:,i])
    # Use parallel processing to get the unique values from the set.
    for i in range(len(xi_map)):
        xi_map[i] = unique(xi_map[i])
    # Return the map and the lookup.
    return xi_map


# Given a categorical input array (either 2D or struct), map this
#  array to an integer encoding matrix with the same number of
#  columns, but unique integers assigned to each unique value.
def i_encode(xi, xi_map):
    xi_rows = xi.shape[0]
    xi_cols = len(xi.dtype) or xi.shape[1]
    _xi = np.zeros((xi_rows, xi_cols), dtype="int64", order="C")
    base = 0
    for i, i_map in enumerate(xi_map):
        # Assign all the integer embeddings.
        values = (xi[:,i] if len(xi.dtype) == 0 else xi[xi.dtype.names[i]])
        if (len(i_map) > 1):
            ix = to_int(values, i_map)
            ix[ix > 0] += base
            _xi[:,i] = ix
            base += len(i_map)
    return _xi


# Convert all inputs to the AXY model into the expected numpy format.
def to_array(ax, axi, sizes, x, xi, y=None, yi=None, yw=None, maps=None):
    if (maps is None):
        maps = dict()
        axi_map = list()
        xi_map = list()
        yi_map = list()
        yi_embeddings = None
    else:
        axi_map = maps.get("axi_map", list())
        xi_map = maps.get("xi_map", list())
        yi_map = maps.get("yi_map", list())
        yi_embeddings = maps.get("yi_embeddings", None)
    # Get the number of inputs.
    if   (y  is not None): nm = len(y)
    elif (yi is not None): nm = len(yi)
    elif (x  is not None): nm = len(x)
    elif (xi is not None): nm = len(xi)
    elif (sizes is not None): nm = len(sizes)
    # Make sure that all inputs are numpy arrays.
    if (y is not None):  y = np.asarray(y, dtype="float32", order="C")
    else:                y = np.zeros((nm,0), dtype="float32", order="C") 
    if (yi is not None): yi = np.asarray(yi)
    else:                yi = np.zeros((nm,0), dtype="int64", order="C")
    if (yw is not None): yw = np.asarray(np.asarray(yw, dtype="float32").reshape((nm,-1)), order="C")
    else:                yw = np.zeros((nm,0), dtype="float32", order="C")
    if (x is not None): x = np.asarray(x, dtype="float32", order="C")
    else:               x = np.zeros((nm,0), dtype="float32", order="C")
    if (xi is not None): xi = np.asarray(xi)
    else:                xi = np.zeros((nm,0), dtype="int64", order="C")
    if (sizes is not None): sizes = np.asarray(sizes, dtype="int64")
    else:                   sizes = np.zeros(0, dtype="int64")
    na = sizes.sum()
    if (ax is not None): ax = np.asarray(ax, dtype="float32", order="C")
    else:                ax = np.zeros((na,0), dtype="float32", order="C")
    if (axi is not None): axi = np.asarray(axi)
    else:                 axi = np.zeros((na,0), dtype="int64", order="C")
    # Make sure that all inputs have the expected shape.
    assert (len(ax.shape) in {1,2}), f"Bad ax shape {ax.shape}, should be 1D or 2D matrix."
    assert (len(axi.shape) in {1,2}), f"Bad axi shape {axi.shape}, should be 1D or 2D matrix."
    assert (len(sizes.shape) == 1), f"Bad sizes shape {sizes.shape}, should be 1D int vector."
    assert (len(x.shape) in {1,2}), f"Bad x shape {x.shape}, should be 1D or 2D matrix."
    assert (len(xi.shape) in {1,2}), f"Bad xi shape {xi.shape}, should be 1D or 2D matrix."
    assert (len(y.shape) in {1,2}), f"Bad y shape {y.shape}, should be 1D or 2D matrix."
    assert (len(yi.shape) in {1,2}), f"Bad yi shape {yi.shape}, should be 1D or 2D matrix."
    assert (len(yw.shape) in {1,2}), f"Bad yw shape {yw.shape}, should be 1D or 2D matrix."
    # Reshape inputs to all be two dimensional (except sizes).
    if (len(ax.shape) == 1): ax = ax.reshape((-1,1))
    if ((len(axi.shape) == 1) and (len(axi.dtype) == 0)): axi = axi.reshape((-1,1))
    if (len(x.shape) == 1): x = x.reshape((-1,1))
    if (len(xi.shape) == 1) and (len(xi.dtype) == 0): xi = xi.reshape((-1,1))
    if (len(y.shape) == 1): y = y.reshape((-1,1))
    if (len(yi.shape) == 1) and (len(yi.dtype) == 0): yi = yi.reshape((-1,1))
    if (len(yw.shape) == 1): yw = yw.reshape((-1,1))
    assert (yw.shape[1] in {0, 1, y.shape[1]}), f"Bad yw shape {yw.shape}, should have 0 columns or 1 column{' or '+str(y.shape[1])+' columns' if (y.shape[1] > 1) else ''}."
    # Set the output size and numeric input sizes.
    mdo = y.shape[1]
    mdn = x.shape[1]
    adn = ax.shape[1]
    # Handle mapping "xi" into integer encodings.
    xi_cols = len(xi.dtype) or xi.shape[1]
    if (xi_cols > 0):
        assert (xi_map is not None), f"Provided data for 'xi' has {xi_cols} columns, 'xi_map' is None."
        if (len(xi_map) == 0):
            xi_map = i_map(xi, xi_map)
        else:
            assert (xi_cols == len(xi_map)), f"Bad number of columns in 'xi', {xi_cols}, expected {len(xi_map)} columns based on provided 'xi_map' map."
        xi = i_encode(xi, xi_map)
        mne = sum(map(len, xi_map))
    else: mne = 0
    # Handle mapping "axi" into integer encodings.
    axi_cols = len(axi.dtype) or axi.shape[1]
    if (axi_cols > 0):
        assert (axi_map is not None), f"Provided data for 'axi' has {axi_cols} columns, 'axi_map' is None."
        if (len(axi_map) == 0):
            axi_map = i_map(axi, axi_map)
        else:
            assert (axi_cols == len(axi_map)), f"Bad number of columns in 'axi', {axi_cols}, expected {len(axi_map)} columns based on provided 'axi_map' map."
        axi = i_encode(axi, axi_map)
        ane = sum(map(len, axi_map))
    else: ane = 0
    # Handle mapping "yi" into integer encodings.
    yi_cols = len(yi.dtype) or yi.shape[1]
    if (yi_cols > 0):
        assert (yi_map is not None), f"Provided data for 'yi' has {yi_cols} columns, 'yi_map' is None."
        if (len(yi_map) == 0):
            yi_map = i_map(yi, yi_map)
        else:
            assert (yi_cols == len(yi_map)), f"Bad number of columns in 'yi', {yi_cols}, expected {len(yi_map)} columns based on provided 'yi_map' map."
        yi = i_encode(yi, yi_map)
        yne = sum(map(len, yi_map))
    else: yne = 0
    # Handle mapping integer encoded "yi" into a single real valued y.
    if (yne > 0):
        # Use a regular simplex to construct equally spaced embeddings for the categories.
        if ((yi_embeddings is None) or (yi_embeddings.size == 0)):
            from tlux.math import regular_simplex
            yi_embeddings = regular_simplex(yne).astype("float32")
        else:
            assert (yi_embeddings.shape[1] == (yne-1)), f"Provided 'yi_embeddings' had shape {yi_embeddings.shape}, but expected a dimension of {yne-1} based on the number of categories."
        # Add a zero vector to the front for "unknown" outputs.
        embedded = np.concatenate((
            np.zeros((1,yne-1), dtype="float32"),
            yi_embeddings), axis=0)
        _y = np.zeros((nm, mdo+yne-1), dtype="float32")
        _y[:,:mdo] = y[:,:]
        for i in range(yi.shape[1]):
            _y[:,mdo:] += embedded[yi[:,i]]
        y = _y
        mdo += yne-1
    # Update all maps and return.
    maps.update(dict(
        axi_map = axi_map,
        xi_map = xi_map,
        yi_map = yi_map,
        yi_embeddings = yi_embeddings,
    ))
    # Return all the shapes and numpy formatted inputs.
    shapes = dict(
        mdn = mdn,
        mne = mne,
        mdo = mdo,
        adn = adn,
        ane = ane,
        yne = yne,
    )
    return nm, na, ax, axi, sizes, x, xi, y, yw, shapes, maps




# Class for timing operations. Initialize to start, call to check, use
# the `start` and `stop` methods / attributes to set / observe values.
class Timer:
    #-----------------------------------------------------------------
    #                            Private
    import time as _time
    _a = 0
    _b = None
    # Offer a function for explicitly beginning a timer.
    def _begin(self):
        self._a = self._time.time()
        self._b = None
        return self._a
    # End function for this timer.
    def _end(self):
        self._b = self._time.time()
        return self._b
    # Return the elapsed time since start.
    def _check(self):
        if   (self._a is None): return 0
        elif (self._b is None): return self._time.time() - self._a
        else:                   return self._b - self._a
    # Overwrite the "__call__" method of the provided object.
    def _callset(obj, func):
        class Float(type(obj)):
            def __call__(self): return func()
            def __repr__(self): return str(float(self))
        return Float(obj)
    #-----------------------------------------------------------------
    #                             Public
    # 
    # If not stopped, return elapsed time, otherwise return total time.
    def __call__(self, precision=2): return float(f"{self._check():.{precision-1}e}")

    # Return "start time" as attribute, "begin timer" as function.
    @property
    def start(self): return Timer._callset(self._a, self._begin)
    # Return "stop time" as attribute, "end timer, return total" as function.
    @property
    def stop(self):
        if (self._b == None): return Timer._callset(self._end(), self.total)
        else:                 return Timer._callset(self._b,     self.total)
    # Return the "total time from start" if running, return "total time" if finished.
    @property
    def total(self):
        if (self._b == None): return Timer._callset(self._check(), self._check)
        else:                 return Timer._callset(self._b - self._a, lambda: self._b - self._a)



import numpy as np
import fmodpy

# Matrix operations module.
random = fmodpy.fimport("../axy_random.f90", dependencies=["pcg32.f90"], name="test_axy_random").random


# --------------------------------------------------------------------
#                            RANDOM_INTEGER

# Make sure that the random integer generation has desired behavior.
def _test_random_integer():
    print("RANDOM_INTEGER..", flush=True)
    counts = {}
    # trials = 100
    trials = 1000000
    bins = 3
    for i in range(trials):
        val = random.random_integer(max_value=bins)
        counts[val] = counts.get(val,0) + 1
    total_count = sum(counts.values())
    for (v,c) in sorted(counts.items()):
        ratio = (c / total_count)
        error = abs(ratio - 1/bins)
        assert (error < 0.001), f"Bad ratio of random integers had value {v} when generating with max_value = {bins}.\n  Ratio was {ratio}\n  Expected {1/bins}"
    print("  passed.", flush=True)


# --------------------------------------------------------------------
#                            RANDOM_REAL

# Make sure that random real number generation has the desired behavior.
def _test_random_real():
    print("RANDOM_REAL..", flush=True)    
    # trials = 10
    trials = 100000
    # 
    # Generate random numbers.
    random.seed_random()
    vals = []
    for i in range(trials):
        vals.append( random.random_real(v=True)[1] )
    # Sort all the random numbers generated.
    vals = sorted(vals)
    # Get the list of linearly spaced real numbers of the same sample size over the unit interval.
    expected = [(i+1)/trials - (1/(2*trials))  for i in range(trials)]
    # Compute the absolute differences between the the sorted values and the evenly spaced ones.
    diffs = [abs(v-e) for (v,e) in zip(vals, expected)]
    # Assert that the maximum difference between the values and a uniform grid is small.
    assert (max(diffs) < 0.01), f"Unexpectedly lumpy distribution of random real numbers for scalar output."
    # 
    # Now repeat, generating all random numbers at once.
    random.seed_random(30183010)
    r = np.zeros((max(1,trials // (20*30)), 20, 30), dtype="float32", order="F")
    random.random_real(r=r, s=r.size)
    vals = sorted(r.flatten())
    expected = [(i+1)/r.size - (1/(2*r.size))  for i in range(r.size)]
    diffs = [abs(v-e) for (v,e) in zip(vals, expected)]
    assert (max(diffs) < 0.01), f"Unexpectedly lumpy distribution of random real numbers for tensor output."
    print("  passed.", flush=True)


# --------------------------------------------------------------------
#                    INDEX_TO_PAIR     PAIR_TO_INDEX

def _test_index_to_pair():
    print("INDEX_TO_PAIR")
    mv = 30
    all_pairs = set()
    # Verify that the pair mapping works forwards and backwards.
    for i in range(1, mv**2+1):
        pair1, pair2 = random.index_to_pair(max_value=mv, i=i)
        all_pairs.add((pair1, pair2))
        j = random.pair_to_index(max_value=mv, pair1=pair1, pair2=pair2)
        assert (i == j), f"Index to pair mapping failed for i={i} max_value={limit} pair={pair} ii={j}."
    # Verify that all pairs were actually generated.
    for i in range(1,mv+1):
        for j in range(1,mv+1):
            assert ((i,j) in all_pairs), f"Pair {(i,j)} missing from enumerated set." 
    print(" passed")


# --------------------------------------------------------------------
#                        INITIALIZE_ITERATOR

def _test_initialize_iterator():
    print("INITIALIZE_ITERATOR..", flush=True)
    nm = 10000
    max_size = 23
    seed = 0
    pairwise = False
    random.seed_random(seed)
    # Initialize space to hold iterators.
    agg_iterators = np.zeros((6,nm), dtype="int64", order="F")    
    # Create sizes.
    sizes = np.asarray(
        [random.random_integer(max_value=max_size) for i in range(nm)],
        dtype="int64", order="F"
    )
    # Initialize all iterators.
    for i in range(nm):
        if (sizes[i] == 0):
            agg_iterators[:,i] = 0
        else:
            agg_iterators[0,i] = sizes[i]
            if pairwise:
                agg_iterators[0,i] = agg_iterators[0,i]**2
            agg_iterators[1:,i] = random.initialize_iterator(
                i_limit=agg_iterators[0,i], seed=seed
            )
    # Check that each aggregator works.
    for i in range(nm):
        if (sizes[i] == 0): continue
        # Generate a full loop of values.
        seen = []
        *agg_iterators[1:,i], next_i = random.get_next_index(*agg_iterators[:,i])
        for _ in range(agg_iterators[0,i]):
            seen.append(next_i)
            *agg_iterators[1:,i], next_i = random.get_next_index(*agg_iterators[:,i], reshuffle=True)
        seen_sorted = tuple( sorted(seen) )
        expected = tuple( range(1,sizes[i]**(2 if pairwise else 1) + 1) )
        assert (seen_sorted == expected), \
            f"Aggregate iterator did not produce expected list of elements.\n  {seen_sorted}\n  {expected}\n  {agg_iterators[:,i]}"
        # Generate another full loop (acknowledging that a reshuffle should have happened).
        past_seen = seen
        seen = []
        *agg_iterators[1:,i], next_i = random.get_next_index(*agg_iterators[:,i])
        for _ in range(agg_iterators[0,i]):
            seen.append(next_i)
            *agg_iterators[1:,i], next_i = random.get_next_index(*agg_iterators[:,i], reshuffle=True)
        seen_sorted = tuple( sorted(seen) )
        assert (seen_sorted == expected), \
            f"Aggregate iterator did not produce expected list of elements.\n  {seen_sorted}\n  {expected}\n  {agg_iterators[:,i]}"
        # Make sure the two loops aren't equal.
        if (sizes[i] > 1):
            assert (tuple(past_seen) != tuple(seen)), f"The two sequences generated were the same.\n  first:  {past_seen}\n  second: {seen}"

    # Done testing.
    print("  passed.", flush=True)


# --------------------------------------------------------------------
#                        RANDOM_UNIT_VECTORS

# TODO: Add test with large number of vectors, ensure no Inf or Nan generated.
def _test_random_unit_vectors():
    print("RANDOM_UNIT_VECTORS..", flush=True)
    # Generate test data.
    n = 5000 # number of points
    for d in (2,): # 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 25, 30, 60, 120):
        bd = 2 # full number of outputs
        br = 1 # reduced rank approximation

        # Random unit vectors.
        a = np.zeros((n,d), dtype="float32")
        random.random_unit_vectors(a.T)

        # Overwrite a with perfectly spaced version for 2 dimensions.
        if (d == 2):
            _ = np.linspace(0, 2*np.pi, n)
            a[:,0] = np.cos(_)
            a[:,1] = np.sin(_)

        # Assert that the distance from the origin for each point is one.
        lengths = np.linalg.norm(a, axis=1)
        length_diff = abs(lengths - 1.0)
        index = np.argmax(length_diff)
        tolerance = 0.0001
        assert (max(length_diff) < tolerance), \
            "One of the vector lengths was significantly different than 1.\n" \
            f"  index = {index}\n" \
            f"  length = {lengths[index]}\n" \
            f"  tolerance = {tolerance}"

        # Assert that the distribution of distances between pairs is not
        #  significantly different than the expected analytic solution for
        #  distances between all pairs of points on a sphere.
        distances = np.concatenate([
            np.linalg.norm(a[:i,:] - a[i], axis=1)
            for i in range(1,len(a))
        ] + [
            np.linalg.norm(a[i+1:,:] - a[i], axis=1)
            for i in range(len(a)-1)
        ], axis=0)


        # If either fail, then plot the vectors.
        from tlux.math import project
        from tlux.plot import Plot, multiplot
        # Plot the actual points (in 3d).
        title = f"3D projection of {d}D points    |    CDF of distances between pairs of {d}D points"
        p1 = Plot(title)
        vecs = project(a, 3)
        if (n <= 100):
            for i in range(n):
                l = np.vstack([[0]*3, vecs[i]])
                p1.add(str(i), *l.T, mode="lines", show_in_legend=False)
        else:
            p1.add("vecs", *vecs.T, marker_size=2, color=1)
        # Plot the distribution of distances.
        p2 = Plot(title, "Distance between pair of points", "P[||p1-p2||] <= Distance")

        # Create the function that converts x and y positions into distances.
        radians = np.linspace(np.pi, 0, 1000)
        x_values = 1 + np.cos(radians)  # [0, ..., 2]
        y_values = np.sin(radians)  # [0, ..., 1, ..., 0]
        dist_values = np.linalg.norm(np.vstack((x_values, y_values)), axis=0)  # [0, ..., 2]

        # def cdf(dist):
        #     # Infer the "x" by pushing the value from the inverse function for distances.
        #     x = np.interp(dist, xp=dist_values, fp=x_values)
        #     # Map the x ∈ [0, 2] to the density.
        #     if (x <= 1): # x ** (1/d) 
        #         return ((1 - (1-x)**(1/(d-1))) / 2)
        #     else:
        #         return ((1/2) + ((x-1) ** (1/(d-1))) / 2)
        # p2.add_func("truth_guess", cdf, [0, 2], color=0, dash="dot")

        p2.add("coordinates", x_values, y_values, mode="lines", color=2)
        p2.add("dist(x)", x_values, dist_values, mode="lines", color=3)
        p2.add("xdist(d)", dist_values, x_values, mode="lines", color=4)

        # Add the distribution of distances.
        cdf_y = np.linspace(0, 1, 1000)
        cdf_x = np.percentile(distances, 100 * cdf_y)
        p2.add("distances CDF", cdf_x, cdf_y, mode="lines", color=1, show_in_legend=False)
        # Show both plots.
        multiplot([p1, p2], show_legend=False)



if __name__ == "__main__":
    print()
    # expected = [121, 72, 47, 110, 69, 116, 59, 90, 17, 32, 71, 70, 93, 76,
    #             83, 50, 41, 120, 95, 30, 117, 36, 107, 10, 65, 80, 119, 118, 13, 3,
    #             98, 89, 40, 15, 78, 37, 84, 27, 58, 113, 39, 38, 61, 44, 51, 18, 9,
    #             88, 63, 85, 4, 75, 106, 33, 48, 87, 86, 109, 92, 99, 66, 57, 8, 111,
    #             46, 5, 52, 26, 81, 96, 7, 6, 29, 12, 19, 114, 105, 56, 31, 94, 53,
    #             100, 43, 74, 1, 16, 55, 54, 77, 60, 67, 34, 25, 104, 79, 14, 101, 20,
    #             91, 49, 64, 103, 102, 108, 115, 82, 73, 24, 62, 21, 68, 11, 42, 97,
    #             112, 23, 22, 45, 28, 35, 2]
    # # Manually check the expected trajectory of an iterator.
    # iterator = [121, 120, 105, 15, 128, 0]
    # # Generate a full loop of values.
    # seen = set()
    # *iterator[1:], next_i = random.get_next_index(*iterator[:], reshuffle=True)
    # for _ in range(iterator[0]+1):
    #     print(next_i, end=", ", flush=True)
    #     seen.add(next_i)
    #     *iterator[1:], next_i = random.get_next_index(*iterator[:], reshuffle=True)
    # seen = tuple( sorted(seen) )
    # expected = tuple( range(1,iterator[0] + 1) )
    # assert (seen == expected), \
    #     f"Aggregate iterator did not produce expected list of elements.\n  {seen}\n  {expected}\n  {agg_iterators[:,i]}"
    # print()
    # exit()

    _test_random_integer()
    _test_random_real()
    _test_index_to_pair()
    _test_initialize_iterator()
    # _test_random_unit_vectors()




# Return a boolean "is_numeric"
def is_numeric(obj):
    try:
        abs((.3*obj + 1*obj) - .3*obj)
        return True
    except: return False

# Store the standard library hash function.
stdlib_hash = hash

# Generate hash values using a near-zero probability of conflict
# hashing mechanism that can handle non-hashable types. Does this
# by hashing the raw bytes (from pickle) with sha256.
def hash(something):
    try:
        return stdlib_hash(something)
    except:
        import hashlib, pickle
        return hashlib.sha256(pickle.dumps(something)).hexdigest()

# Generate a sorted unique set of values from an iterable, if the
# values are not hashable, use a hashing function to sort them.
def sorted_unique(iterable):
    try:
        # Try returning sorted unique values using hash to find unique.
        return sorted(set(iterable))
    except TypeError: 
        # Get the set of unique hashes and associated values.
        ids = {hash(v):v for v in iterable}
        # Try sorting the values directly, if that's not possible
        # then sort the values by their hash.
        try:    return sorted(ids.values())
        except: return [ids[h] for h in sorted(ids)]


# Given "d" categories that need to be converted into real space,
# generate a regular simplex in (d-1)-dimensional space. (all points
# are equally spaced from each other and the origin) This process
# guarantees that all points are not placed on a sub-dimensional
# manifold (as opposed to "one-hot" encoding).
def regular_simplex(num_categories):
    import numpy as np
    class InvalidNumberOfCategories(Exception): pass
    # Special cases for one and two categories
    if num_categories < 1:
        raise(InvalidNumberOfCategories(
            "Number of categories must be an integer greater than 0."))
    elif num_categories == 1:
        return np.array([[]])
    elif num_categories == 2:
        return np.array([[0.],[1.]])
    # Standard case for >2 categories
    d = num_categories
    # Initialize all points to be zeros.
    points = np.zeros((d,d-1))
    # Set the initial first point as 1
    points[0,0] = 1
    # Calculate all the intermediate points
    for i in range(1,d-1):
        # Set all points to be flipped from previous calculation while
        # maintaining the angle "arcos(-1/d)" between vectors.
        points[i:,i-1] = -1/(d-i) * points[i-1,i-1]
        # Compute the new coordinate using pythagorean theorem
        points[i,i] = (1 - sum(points[i,:i]**2))**(1/2)
    # Set the last coordinate of the last point as the negation of the previous
    points[i+1,i] = -points[i,i]
    # Return the regular simplex
    return points


# Given a point, determine the affine combination of categories in the
# defining regular simplex that produces the point.
def category_ratio(point):
    import numpy as np
    categories = regular_simplex(len(point)+1)
    # Add the "constant" column to the matrix of categories
    categories = np.hstack((categories, np.ones((len(point)+1,1))))
    # Add the "sum to 1" column to the affinely created point
    point = np.hstack((point, 1))
    # Calculate and return the affine weights
    return np.linalg.solve(categories.T, point)


# Return a randomized "range" using the appropriate technique based on
# the size of the range being generated. If the memory footprint is
# small (<= 32K int's) then a random sample is created and returned.
# If the memory footprint would be large, a Linear Congruential
# Generator is used to efficiently generate the sequence.
# 
# Parameters are similar to the builtin `range` with:
#   start -- int, default of 0.
#   stop  -- int > start for no / positive step, < start otherwise.
#   step  -- int (!= 0), default of 1.
#   count -- int > 0, number of samples, default (stop-start)//step.
# 
# Usage (and implied parameter ordering):
#   random_range(a)             --> range(0, a, 1)[:a]
#   random_range(a, b) [a < b]  --> range(a, b, 1)[:b-a]
#   random_range(a, b, c)       --> range(a, b, c)[:(b-a)//c]
#   random_range(a, b, c, d)    --> range(a, b, c)[:min(d, (b-a)//c)]
#   random_range(a, d) [d <= a] --> range(0, a, 1)[:d]
# 
# If the size of the range is large, a Linear Congruential Generator is used.
#   Memory  -- O(1) storage for a few integers, regardless of parameters.
#   Compute -- O(n) at most 2 times the number of steps in the range, n.
# 
def random_range(start, stop=None, step=None, count=float('inf')):
    from random import sample, randint
    from math import ceil, log2
    # Add special usage where the second argument is meant to be a count.
    if (stop != None) and (stop <= start) and ((step == None) or (step >= 0)):
        start, stop, count = 0, start, stop
    # Set a default values the same way "range" does.
    if (stop == None): start, stop = 0, start
    if (step == None): step = 1
    # Compute the number of numbers in this range, update count accordingly.
    num_steps = int((stop - start) // step)
    count = min(count, num_steps)
    # Check for a usage error.
    if (num_steps == 0) or (count <= 0): raise(InvalidRange(start, stop, step, count))
    # Use robust random method if it has a small enough memory footprint.
    # Do not use this method if there is a number too large for C implementation.
    if (num_steps <= 2**15) and (max(map(abs,(start,stop,step))) < 2**63):
        for value in sample(range(start,stop,step), count): yield value
        return
    # Use the LCG for the cases where the above is too memory intensive.
    # Enforce that all numbers are python integers (no worry of overflow).
    start, stop, step = map(int, (start, stop, step))
    # Use a mapping to convert a standard range into the desired range.
    mapping = lambda i: (i*step) + start
    # Seed range with a random integer to start.
    value = int(randint(0,num_steps))
    # 
    # Construct an offset, multiplier, and modulus for a linear
    # congruential generator. These generators are cyclic and
    # non-repeating when they maintain the properties:
    # 
    #   1) "modulus" and "offset" are relatively prime.
    #   2) ["multiplier" - 1] is divisible by all prime factors of "modulus".
    #   3) ["multiplier" - 1] is divisible by 4 if "modulus" is divisible by 4.
    # 
    offset = int(randint(0,num_steps)) * 2 + 1                 # Pick a random odd-valued offset.
    multiplier = 4*(num_steps + int(randint(0,num_steps))) + 1 # Pick a multiplier 1 greater than a multiple of 4.
    modulus = int(2**ceil(log2(num_steps)))                    # Pick a modulus just big enough to generate all numbers (power of 2).
    # Track how many random numbers have been returned.
    found = 0
    while found < count:
        # If this is a valid value, yield it in generator fashion.
        if value < num_steps:
            found += 1
            yield mapping(value)
        # Calculate the next value in the sequence.
        value = (value*multiplier + offset) % modulus


# Calculate the maximum difference between two CDF functions (two sample).
def ks_diff(l1, l2, n=100):
    import numpy as np
    # Sort both sets of points.
    l1 = sorted(l1)
    l2 = sorted(l2)
    # Get all unique points sorted.
    union = sorted(set(l1) | set(l2))
    # Compute the max difference between empirical distribution function values.
    max_diff = 0
    for v in union:
        p1_cdf = np.searchsorted(l1, v, side='right') / len(l1)
        p2_cdf = np.searchsorted(l2, v, side='right') / len(l2)
        max_diff = max(max_diff, abs(p1_cdf - p2_cdf))
    return max_diff


# Given a ks-statistic and the sample sizes of the two distributions
# compared, return the largest confidence with which the two
# distributions can be said to be the same.
def ks_p_value(ks_stat, n1, n2=float('inf')):
    # By definition of the KS-test: 
    # 
    #    KS > c(a) (1/n1 + 1/n2)^(1/2)
    # 
    #   where KS is the KS statistic, n1 and n2 are the respective
    #   sample sizes for each distribution and c(a) is defined as
    # 
    #    c(a) = ( -ln(a/2)/2 )^(1/2)
    # 
    #   is the standard for testing the probability with which two
    #   distributions come from different underlying distributions. If
    #   we want the distributions to be the same, we want the KS test
    #   to only pass with large values for "a" (large 'p-value'). The
    #   above check can be reversed to compute "a", which provides the
    #   largest p-value for which the KS test states the two
    #   distributions are not certainly different.
    # 
    #    c^-1(b) = 2 e^(-2 b^2)
    # 
    #    a = c^-1 ( (KS / (1/n1 + 1/n2)^(1/2)) )
    #      = 2 e^( -2 (KS / (1/n1 + 1/n2)^(1/2))^2 )
    #      = 2 e^( -2 KS^2 / |1/n1 + 1/n2| )
    #    
    #    and "a" cannot be larger than 1. Therefore, finally we have
    # 
    #    a = min(1, 2 e^( -2 KS^2 / |1/n1 + 1/n2| ))
    #                                                                 
    return min(1.0, 2 * np.exp( -2 * ( ks_stat**2 / abs((1/n1) + (1/n2)) )))


# Use a linear interpolation of the EDF points to compute the integral
# of the absolute difference between the empirical PDF's of two sequences.
def epdf_diff(seq_1, seq_2):
    # Construct generators for the EDF point pairs.
    gen_1 = edf_pair_gen(seq_1)
    gen_2 = edf_pair_gen(seq_2)
    # Get the initial values from the generators.
    low_1, upp_1, density_1 = gen_1.__next__()
    low_2, upp_2, density_2 = gen_2.__next__()
    shared = 0.
    # Cycle until completing the integral difference between pEDFs.
    while (upp_1 < float('inf')) or (upp_2 < float('inf')):
        # Compute the width of the interval and the percentage of data
        # within the width for each of the distributions.
        width = min(upp_1, upp_2) - max(low_1, low_2)
        if width < float('inf'):
            # Convert the EDF points into densities over the interval.
            sub_density_1 = density_1 * (width / (upp_1-low_1))
            sub_density_2 = density_2 * (width / (upp_2-low_2))
            if abs(sub_density_1 - sub_density_2) >= 0:
                shared += min(sub_density_1, sub_density_2)
        # Cycle whichever range has the lower maximum.
        if (upp_1 > upp_2):
            low_2, upp_2, density_2 = next(gen_2)
        elif (upp_2 > upp_1):
            low_1, upp_1, density_1 = next(gen_1)
        else:
            low_1, upp_1, density_1 = next(gen_1)
            low_2, upp_2, density_2 = next(gen_2)
    return max(0, 1 - shared)


# Compute the effect between two sequences. Use the following:
#    number vs. number     -- Correlation coefficient between the two sequences.
#    category vs. number   -- "method" L1-norm difference between full distribution
#                             and conditional distributions for categories.
#    category vs. category -- "method" total difference between full distribution
#                             of other sequence given value of one sequence.
# 
# INPUTS:
# 
#   seq_1  -- An iterable of objects with a "__len__" property.
#   seq_2  -- An iterable of objects with a "__len__" property.
#   method -- Either "max", "min", "mean" or None. Determines how the
#             "effect" is computed for cat-num and cat-cat computations,
#             since each will yield multiple difference values.
#             (method==None) returns the full set of differences.
#   use_ks -- True or False, when measuring cat-num effect, if True, 
#             use the KS test p-value between conditional distributions
#             to compute effect, if Falsd use the e-pdf L1 difference.
# 
def effect(seq_1, seq_2, method="mean", use_ks=False):
    import numpy as np
    # Check for equally lengthed sequences.
    class CannotCompare(Exception): pass
    if len(seq_1) != len(seq_2): raise(CannotCompare("Provided sequences must have same length for comparison to be made."))
    # Check for index accessibility.
    try:              seq_1[0], seq_2[0]
    except TypeError: raise(CannotCompare("Provided sequences must both be index-accessible."))
    # Get the set of attributes a numeric type should have.
    # The three different outcomes of comparison could be:
    num_num = (is_numeric(seq_1[0]) and is_numeric(seq_2[0]))
    cat_cat = (not is_numeric(seq_1[0])) and (not is_numeric(seq_2[0]))
    num_cat = (not cat_cat) and (not num_num)
    # Execute the appropriate comparison according to the types.
    if num_num:
        # Compute the correlation (highly correlated have no difference).
        return float(np.corrcoef(seq_1, seq_2)[0,1])
    elif num_cat:
        # Compute the change in numeric distribution caused by a chosen category.
        nums, cats = (seq_1, seq_2) if is_numeric(seq_1[0]) else (seq_2, seq_1)
        # Compute the sorted order for "nums" by index.
        sort = sorted(range(len(nums)), key=lambda i: nums[i])
        nums, cats = [nums[i] for i in sort], [cats[i] for i in sort]
        # Consider each unique categorical value, compute epdf diff.
        diffs = {}
        for cat in sorted_unique(cats):
            main_nums = [n for (n,c) in zip(nums,cats) if c != cat]
            sub_nums = [n for (n,c) in zip(nums,cats) if c == cat]
            if (len(main_nums) == 0) or (len(sub_nums) == 0):
                dist_diff = 0.0
            elif (not use_ks):
                dist_diff = epdf_diff(main_nums, sub_nums)
            else:
                # Use the KS-statistic to test the likelihood that the two
                # sets come from different underlying distributions.
                dist_diff = 1.0 - ks_p_value(ks_diff(main_nums, sub_nums), 
                                             len(main_nums), len(sub_nums))
            try:              diffs[cat]       = dist_diff
            except TypeError: diffs[hash(cat)] = dist_diff
    elif cat_cat:
        diffs = {}
        # Generate unique lists of values (hash those that aren't).
        unique_1 = sorted_unique(seq_1)
        unique_2 = sorted_unique(seq_2)
        # Compute the percentage of each unique category for full sequences.
        full_1 = categorical_pdf(seq_1, unique_1)
        full_2 = categorical_pdf(seq_2, unique_2)
        # Identify the difference when selecting sequence 1 values.
        for cat in unique_1:
            main_seq = categorical_pdf([c2 for (c1,c2) in zip(seq_1,seq_2) if c1 != cat])
            sub_seq = categorical_pdf([c2 for (c1,c2) in zip(seq_1,seq_2) if c1 == cat])
            if (len(main_seq) > 0) and (len(sub_seq) > 0):
                dist_diff = categorical_diff( main_seq, sub_seq )
            else: dist_diff = 0.
            try:              diffs[(cat,None)]       = dist_diff
            except TypeError: diffs[(hash(cat),None)] = dist_diff
        # Identify the difference when selecting sequence 2 values.
        for cat in unique_2:
            main_seq = categorical_pdf([c1 for (c1,c2) in zip(seq_1,seq_2) if c2 != cat])
            sub_seq = categorical_pdf([c1 for (c1,c2) in zip(seq_1,seq_2) if c2 == cat])
            if (len(main_seq) > 0) and (len(sub_seq) > 0):
                dist_diff = categorical_diff( main_seq, sub_seq )
            else: dist_diff = 0.
            try:              diffs[(None,cat)]       = dist_diff
            except TypeError: diffs[(None,hash(cat))] = dist_diff
    # Return the desired measure of difference between the two.
    if   (method == "max"):  return max(diffs.values())
    elif (method == "min"):  return min(diffs.values())
    elif (method == "mean"): return sum(diffs.values()) / len(diffs)
    else:                    return diffs




import time
import numpy as np
import fmodpy

# Import two different linear models.
from tlux.approximate.apos import APOS

# Create data.
np.random.seed(0)
n = 100
d = 6

# Create test data.
x = np.random.random(size=(n,d)).astype("float32")
y = np.cos(x * np.linspace(1,d,d))[:,:(d+1)//2].astype("float32")

print("x.shape: ", x.shape)
print("y.shape: ", y.shape)

# Fit the APOS model.
alin = APOS(mdn=x.shape[1], mdo=y.shape[1], adn=0, mns=0)
_ = time.time()
alin.fit(x, y, steps=10)
alin_total = time.time() - _
print("alin_total: ", alin_total)

# Fit the linear model.
_ = time.time()
model, residuals = np.linalg.lstsq(x, y, rcond=None)[:2]
lin_total = time.time() - _
print(" lin_total: ", lin_total)
print(model)

lp = fmodpy.fimport("test_sgels.f90", lapack=True, omp=False).least_squares
xf = np.asarray(x.T, order="F")
yf = np.asarray(y, order="F")
_ = time.time()
s = lp(bytes('T','ASCII'), xf, yf)[-1]
lp_total = time.time() - _
print("  lp_total: ", lp_total)
print("s.shape: ", s.shape)
print(s)

# TODO:
#  - compute useless nodes in a layer by orthogonalization
#  - compute output values of noncontributors, least squares fit
#    the values that are kept to those noncontributing outputs
#  - update the weights of the remaining nodes and delete useless ones
#  - reduce the gradient afterwards to its principal components that
#    fill the space of values
#  - least squares fit the data at the layer before to the gradient
#    placed in the previously useless values
#  - orthogonalize all new values to see if rank has improved
#  - fill remaining empty rank with first K principal components of
#    preceeding layer
#  - pick shift terms for all new nodes as the value that maximizes
#    alignment with the gradient (first), and capturing all previous
#    layer values (principal components) second (shift = minval),
#    and increasing diversity of output (distance from neighbor
#    states) third
# 




# Function for printing the help message.
def print_help_message():
    print("""

balltree -- A very fast python-wrapped fortran implementation of a
    `ball tree`, includes fast `argsort` and `argselect` codes for
    flat arrays as well.

### Python

```python
import numpy as np
import balltree

points = np.random.random(size=(100000,1000))
tree = balltree.BallTree(points, [leaf_size=<leaf size int>])
distances, indices = tree.query(points[:1,:], [k=<k nearest int>])
# 'distances' will have the distances to the `k` nearest points to points[0]
# 'indices' will have the index of the `k` nearest points to points[0]


x = np.random.random(size=(10000,))
x, i = balltree.argsort(x, i=np.arange(len(x)))
# `x` is sorted in place
# `i` is sorted symmetrically with `x` in place, in this case it
#     shows the original indices of the elements of `x` (before sort).


x = np.random.random(size=(10000,))
k = 0
x, i = balltree.argselect(x, k, i=np.arange(len(x)))
# `x` is rearranged such that the element at `x[k]` has rank `k`
# `i` is rearranged symmetrically with `x` in place, in this case it
#     shows the original indices of the elements of `x` (before sort).
```

    """)

# If no arguments were provided, give the usage.
print_help_message()




'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.

! Module for all random number generation used in the AXY package.
'''

import os
import ctypes
import platform
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "axy_random." + platform.machine() + ".so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-fPIC', '-shared', '-O3']
_ordered_dependencies = ['pcg32.f90', 'axy_random.f90', 'axy_random_c_wrapper.f90']
_symbol_files = []# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the prerequisite symbols for the compiled code.
for _ in _symbol_files:
    _ = ctypes.CDLL(os.path.join(_this_directory, _), mode=ctypes.RTLD_GLOBAL)
# Try to import the existing object. If that fails, recompile and then try.
try:
    # Check to see if the source files have been modified and a recompilation is needed.
    if (max(max([0]+[os.path.getmtime(os.path.realpath(os.path.join(_this_directory,_))) for _ in _symbol_files]),
            max([0]+[os.path.getmtime(os.path.realpath(os.path.join(_this_directory,_))) for _ in _ordered_dependencies]))
        > os.path.getmtime(_path_to_lib)):
        print()
        print("WARNING: Recompiling because the modification time of a source file is newer than the library.", flush=True)
        print()
        if os.path.exists(_path_to_lib):
            os.remove(_path_to_lib)
        raise NotImplementedError(f"The newest library code has not been compiled.")
    # Import the library.
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


class random:
    ''''''

    # Declare 'zero'
    def get_zero(self):
        zero = ctypes.c_long()
        clib.random_get_zero(ctypes.byref(zero))
        return zero.value
    def set_zero(self, zero):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    zero = property(get_zero, set_zero)

    # Declare 'one'
    def get_one(self):
        one = ctypes.c_long()
        clib.random_get_one(ctypes.byref(one))
        return one.value
    def set_one(self, one):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    one = property(get_one, set_one)

    # Declare 'two'
    def get_two(self):
        two = ctypes.c_long()
        clib.random_get_two(ctypes.byref(two))
        return two.value
    def set_two(self, two):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    two = property(get_two, set_two)

    # Declare 'four'
    def get_four(self):
        four = ctypes.c_long()
        clib.random_get_four(ctypes.byref(four))
        return four.value
    def set_four(self, four):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    four = property(get_four, set_four)

    # Declare 'pi'
    def get_pi(self):
        pi = ctypes.c_float()
        clib.random_get_pi(ctypes.byref(pi))
        return pi.value
    def set_pi(self, pi):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    pi = property(get_pi, set_pi)

    # Declare 'right_32'
    def get_right_32(self):
        right_32 = ctypes.c_long()
        clib.random_get_right_32(ctypes.byref(right_32))
        return right_32.value
    def set_right_32(self, right_32):
        raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))
    right_32 = property(get_right_32, set_right_32)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine SEED_RANDOM
    
    def seed_random(self, seed=None):
        '''! Set the seed for the random number generator.'''
        
        # Setting up "seed"
        seed_present = ctypes.c_bool(True)
        if (seed is None):
            seed_present = ctypes.c_bool(False)
            seed = ctypes.c_long()
        else:
            seed = ctypes.c_long(seed)
        if (type(seed) is not ctypes.c_long): seed = ctypes.c_long(seed)
    
        # Call C-accessible Fortran wrapper.
        clib.c_seed_random(ctypes.byref(seed_present), ctypes.byref(seed))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return 

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine RANDOM_REAL
    
    def random_real(self, r=None, s=None, v=None):
        '''! Fortran implementation of the algorithm for generating uniform likelihood
!  real numbers described at http://mumble.net/~campbell/tmp/random_real.c'''
        
        # Setting up "r"
        r_present = ctypes.c_bool(True)
        if (r is None):
            r_present = ctypes.c_bool(False)
            r = numpy.zeros(shape=(1), dtype=ctypes.c_float, order='F')
        elif (type(r) == bool) and (r):
            r = numpy.zeros(shape=(1), dtype=ctypes.c_float, order='F')
        elif ((not issubclass(type(r), numpy.ndarray)) or
              (not numpy.asarray(r).flags.f_contiguous) or
              (not (r.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'r' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            r = numpy.asarray(r, dtype=ctypes.c_float, order='F')
        if (r_present):
            r_dim_1 = ctypes.c_long(r.shape[0])
        else:
            r_dim_1 = ctypes.c_long()
        
        # Setting up "s"
        s_present = ctypes.c_bool(True)
        if (s is None):
            s_present = ctypes.c_bool(False)
            s = ctypes.c_long()
        else:
            s = ctypes.c_long(s)
        if (type(s) is not ctypes.c_long): s = ctypes.c_long(s)
        
        # Setting up "v"
        v_present = ctypes.c_bool(True)
        if (v is None):
            v_present = ctypes.c_bool(False)
            v = ctypes.c_float()
        else:
            v = ctypes.c_float(v)
    
        # Call C-accessible Fortran wrapper.
        clib.c_random_real(ctypes.byref(r_present), ctypes.byref(r_dim_1), ctypes.c_void_p(r.ctypes.data), ctypes.byref(s_present), ctypes.byref(s), ctypes.byref(v_present), ctypes.byref(v))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return (r if r_present else None), (v.value if v_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine RANDOM_UNIT_VECTORS
    
    def random_unit_vectors(self, column_vectors):
        '''! Generate randomly distributed vectors on the N-sphere.'''
        
        # Setting up "column_vectors"
        if ((not issubclass(type(column_vectors), numpy.ndarray)) or
            (not numpy.asarray(column_vectors).flags.f_contiguous) or
            (not (column_vectors.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'column_vectors' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            column_vectors = numpy.asarray(column_vectors, dtype=ctypes.c_float, order='F')
        column_vectors_dim_1 = ctypes.c_long(column_vectors.shape[0])
        column_vectors_dim_2 = ctypes.c_long(column_vectors.shape[1])
    
        # Call C-accessible Fortran wrapper.
        clib.c_random_unit_vectors(ctypes.byref(column_vectors_dim_1), ctypes.byref(column_vectors_dim_2), ctypes.c_void_p(column_vectors.ctypes.data))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return column_vectors

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine INITIALIZE_ITERATOR
    
    def initialize_iterator(self, i_limit, seed=None):
        '''! Given the variables for a linear iterator, initialize it.'''
        
        # Setting up "i_limit"
        if (type(i_limit) is not ctypes.c_long): i_limit = ctypes.c_long(i_limit)
        
        # Setting up "i_next"
        i_next = ctypes.c_long()
        
        # Setting up "i_mult"
        i_mult = ctypes.c_long()
        
        # Setting up "i_step"
        i_step = ctypes.c_long()
        
        # Setting up "i_mod"
        i_mod = ctypes.c_long()
        
        # Setting up "i_iter"
        i_iter = ctypes.c_long()
        
        # Setting up "seed"
        seed_present = ctypes.c_bool(True)
        if (seed is None):
            seed_present = ctypes.c_bool(False)
            seed = ctypes.c_long()
        else:
            seed = ctypes.c_long(seed)
        if (type(seed) is not ctypes.c_long): seed = ctypes.c_long(seed)
    
        # Call C-accessible Fortran wrapper.
        clib.c_initialize_iterator(ctypes.byref(i_limit), ctypes.byref(i_next), ctypes.byref(i_mult), ctypes.byref(i_step), ctypes.byref(i_mod), ctypes.byref(i_iter), ctypes.byref(seed_present), ctypes.byref(seed))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return i_next.value, i_mult.value, i_step.value, i_mod.value, i_iter.value

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine INDEX_TO_PAIR
    
    def index_to_pair(self, max_value, i):
        '''! Map an integer I in the range [1, MAX_VALUE**2] to a unique pair
!  of integers PAIR1 and PAIR2 with both in the range [1, MAX_VALUE].'''
        
        # Setting up "max_value"
        if (type(max_value) is not ctypes.c_long): max_value = ctypes.c_long(max_value)
        
        # Setting up "i"
        if (type(i) is not ctypes.c_long): i = ctypes.c_long(i)
        
        # Setting up "pair1"
        pair1 = ctypes.c_long()
        
        # Setting up "pair2"
        pair2 = ctypes.c_long()
    
        # Call C-accessible Fortran wrapper.
        clib.c_index_to_pair(ctypes.byref(max_value), ctypes.byref(i), ctypes.byref(pair1), ctypes.byref(pair2))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return pair1.value, pair2.value

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine PAIR_TO_INDEX
    
    def pair_to_index(self, max_value, pair1, pair2):
        '''! Map a pair of integers PAIR1 and PAIR2 in the range [1, MAX_VALUE]
!  to an integer I in the range [1, MAX_VALUE**2].'''
        
        # Setting up "max_value"
        if (type(max_value) is not ctypes.c_long): max_value = ctypes.c_long(max_value)
        
        # Setting up "pair1"
        if (type(pair1) is not ctypes.c_long): pair1 = ctypes.c_long(pair1)
        
        # Setting up "pair2"
        if (type(pair2) is not ctypes.c_long): pair2 = ctypes.c_long(pair2)
        
        # Setting up "i"
        i = ctypes.c_long()
    
        # Call C-accessible Fortran wrapper.
        clib.c_pair_to_index(ctypes.byref(max_value), ctypes.byref(pair1), ctypes.byref(pair2), ctypes.byref(i))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return i.value

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine RANDOM_INTEGER
    
    def random_integer(self, max_value=None):
        '''! Define a function for generating random integers.
! Optional MAX_VALUE is a noninclusive upper bound for the value generated.
!   WARNING: Only generates integers in the range [0, 2**32).'''
        
        # Setting up "max_value"
        max_value_present = ctypes.c_bool(True)
        if (max_value is None):
            max_value_present = ctypes.c_bool(False)
            max_value = ctypes.c_long()
        else:
            max_value = ctypes.c_long(max_value)
        if (type(max_value) is not ctypes.c_long): max_value = ctypes.c_long(max_value)
        
        # Setting up "random_int"
        random_int = ctypes.c_long()
    
        # Call C-accessible Fortran wrapper.
        clib.c_random_integer(ctypes.byref(max_value_present), ctypes.byref(max_value), ctypes.byref(random_int))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return random_int.value

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine GET_NEXT_INDEX
    
    def get_next_index(self, i_limit, i_next, i_mult, i_step, i_mod, i_iter, reshuffle=None):
        '''! Get the next index in the model point iterator.'''
        
        # Setting up "i_limit"
        if (type(i_limit) is not ctypes.c_long): i_limit = ctypes.c_long(i_limit)
        
        # Setting up "i_next"
        if (type(i_next) is not ctypes.c_long): i_next = ctypes.c_long(i_next)
        
        # Setting up "i_mult"
        if (type(i_mult) is not ctypes.c_long): i_mult = ctypes.c_long(i_mult)
        
        # Setting up "i_step"
        if (type(i_step) is not ctypes.c_long): i_step = ctypes.c_long(i_step)
        
        # Setting up "i_mod"
        if (type(i_mod) is not ctypes.c_long): i_mod = ctypes.c_long(i_mod)
        
        # Setting up "i_iter"
        if (type(i_iter) is not ctypes.c_long): i_iter = ctypes.c_long(i_iter)
        
        # Setting up "reshuffle"
        reshuffle_present = ctypes.c_bool(True)
        if (reshuffle is None):
            reshuffle_present = ctypes.c_bool(False)
            reshuffle = ctypes.c_bool()
        else:
            reshuffle = ctypes.c_bool(reshuffle)
        if (type(reshuffle) is not ctypes.c_bool): reshuffle = ctypes.c_bool(reshuffle)
        
        # Setting up "next_i"
        next_i = ctypes.c_long()
    
        # Call C-accessible Fortran wrapper.
        clib.c_get_next_index(ctypes.byref(i_limit), ctypes.byref(i_next), ctypes.byref(i_mult), ctypes.byref(i_step), ctypes.byref(i_mod), ctypes.byref(i_iter), ctypes.byref(reshuffle_present), ctypes.byref(reshuffle), ctypes.byref(next_i))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return i_next.value, i_mult.value, i_step.value, i_mod.value, i_iter.value, next_i.value

random = random()






# Build a class that contains pointers to the model internals, allowing
#  python attribute access to all of the different components of the models.
class AxyModel:
    def __init__(self, config, model, clock_rate=2000000000, show_vecs=False, show_times=True):
        self.config = config
        self.model = model
        self.clock_rate = clock_rate
        self.show_vecs = show_vecs
        self.show_times = show_times
        self.ax_shift      = self.model[self.config.aiss-1:self.config.aise].reshape(self.config.adn, order="F")
        self.ax_rescale    = self.model[self.config.aims-1:self.config.aime].reshape(self.config.adn, self.config.adn, order="F")
        self.a_embeddings  = self.model[self.config.asev-1:self.config.aeev].reshape(self.config.ade, self.config.ane, order="F")
        self.a_input_vecs  = self.model[self.config.asiv-1:self.config.aeiv].reshape(self.config.adi, self.config.ads, order="F")
        self.a_input_shift = self.model[self.config.asis-1:self.config.aeis].reshape(self.config.ads, order="F")
        self.a_state_vecs  = self.model[self.config.assv-1:self.config.aesv].reshape(self.config.ads, self.config.ads, max(0,self.config.ans-1), order="F")
        self.a_state_shift = self.model[self.config.asss-1:self.config.aess].reshape(self.config.ads, max(0,self.config.ans-1), order="F")
        self.a_output_vecs = self.model[self.config.asov-1:self.config.aeov].reshape(self.config.adso, self.config.ado+1, order="F")
        self.ay_shift      = self.model[self.config.aoss-1:self.config.aose].reshape(self.config.ado, order="F")
        self.x_shift       = self.model[self.config.miss-1:self.config.mise].reshape(self.config.mdn, order="F")
        self.x_rescale     = self.model[self.config.mims-1:self.config.mime].reshape(self.config.mdn, self.config.mdn, order="F")
        self.m_embeddings  = self.model[self.config.msev-1:self.config.meev].reshape(self.config.mde, self.config.mne, order="F")
        self.m_input_vecs  = self.model[self.config.msiv-1:self.config.meiv].reshape(self.config.mdi, self.config.mds, order="F")
        self.m_input_shift = self.model[self.config.msis-1:self.config.meis].reshape(self.config.mds, order="F")
        self.m_state_vecs  = self.model[self.config.mssv-1:self.config.mesv].reshape(self.config.mds, self.config.mds, max(0,self.config.mns-1), order="F")
        self.m_state_shift = self.model[self.config.msss-1:self.config.mess].reshape(self.config.mds, max(0,self.config.mns-1), order="F")
        self.m_output_vecs = self.model[self.config.msov-1:self.config.meov].reshape(self.config.mdso, self.config.mdo, order="F")
        self.y_shift       = self.model[self.config.moss-1:self.config.mose].reshape(self.config.do, order="F")
        self.y_rescale     = self.model[self.config.moms-1:self.config.mome].reshape(self.config.do, self.config.do, order="F")

    # Allow square brackets to access attributes of this model and its configuration.
    def __getitem__(self, attr):
        if hasattr(self.config, attr):
            return getattr(self.config, attr)
        else:
            return self.__getattribute__(attr)

    # Allow the "." operator to access attributes of this model and its configuration.
    def __getattr__(self, attr):
        if hasattr(self.config, attr):
            return getattr(self.config, attr)
        else:
            return self.__getattribute__(attr)

    # Create a string summarizing how time has been spent for this model.
    def _timer_summary_string(self):
        config = self.config
        clock_rate = self.clock_rate
        # Show the summary of time spent on different operations.
        time_vars = [
            ("int", "initialize"),
            ("fit", "fit"),
            ("nrm", "normalize"),
            ("gen", "fetch data"),
            ("emb", "embed"),
            ("evl", "evaluate"),
            ("grd", "gradient"),
            ("rat", "rate update"),
            ("opt", "step vars"),
            ("con", "condition"),
            ("enc", "encode")
        ]
        max_p, max_n = max(time_vars, key=lambda pn: getattr(config, "w"+pn[0]))
        total = getattr(config, "w"+max_p)
        if (total <= 0): return ""
        max_len = max((len(n) for (p,n) in time_vars))
        # Generate a summary line for each item.
        time_str = f" time category    sec {max_n+'%':>12s}   speedup\n"
        for (p, n) in time_vars:
            v = getattr(config, "w"+p)
            pv = getattr(config, "c"+p)
            time_str += f"  {n:{max_len}s}   {v/(clock_rate if clock_rate > 0 else 1):.1e}s   ({100.0*v/(total if total > 0 else 1):5.1f}%)   [{pv/(v if v > 0 else 1):.1f}x]\n"
        return "\n" + time_str + "\n"

    # Create a summary string for this model.
    def __str__(self, show_vecs=False, show_times=False):
        show_vecs = show_vecs or self.show_vecs
        # A function for creating a byte-size string from an integer.
        def _byte_str(byte_size):
            if (byte_size < 2**10):
                byte_size = f"{byte_size} bytes"
            elif (byte_size < 2**20):
                byte_size = f"{byte_size//2**10:.1f}KB"
            elif (byte_size < 2**30):
                byte_size = f"{byte_size//2**20:.1f}MB"
            elif (byte_size < 2**40):
                byte_size = f"{byte_size//2**30:.1f}GB"
            else:
                byte_size = f"{byte_size//2**40:.1f}TB"
            return byte_size
        # Calculate the byte size of this model (excluding python descriptors).
        # TODO: Not all configs are 4 bytes, do more expensive sum over actual sizes?
        byte_size = len(self.config._fields_)*4 + self.model.dtype.itemsize*self.model.size
        byte_size = _byte_str(byte_size)
        if (self.config.rwork_size+self.config.iwork_size > 0):
            work_size = self.config.rwork_size*4 + self.config.iwork_size*4
            byte_size += " + "+_byte_str(work_size)+" work space"
        # Create a function that prints the actual contents of the arrays.
        if show_vecs: to_str = lambda arr: ("\n    " + "\n    ".join(str(arr).split("\n")) + "\n" if (arr.size > 0) else "\n")
        else:         to_str = lambda arr: "\n"
        # Provide details (and some values where possible).
        return (
            f"AXY model ({self.config.num_vars} variables, {self.config.total_size-self.config.num_vars} parameters) [{byte_size}]\n"+
            (" aggregator model\n"+
            f"  input dimension  {self.config.adn}{' + '+str(self.config.ade) if self.config.ade > 0 else ''}\n"+
            f"  output dimension {self.config.ado}\n"+
            f"  state dimension  {self.config.ads}\n"+
            f"  number of states {self.config.ans}\n"+
           (f"  embedding dimension  {self.config.ade}\n"+
            f"  number of embeddings {self.config.ane}\n"
             if self.config.ane > 0 else "")+
            f"  embeddings   {self.a_embeddings.shape}  "+to_str(self.a_embeddings)+
            f"  input vecs   {self.a_input_vecs.shape}  "+to_str(self.a_input_vecs)+
            f"  input shift  {self.a_input_shift.shape} "+to_str(self.a_input_shift)+
            f"  state vecs   {self.a_state_vecs.shape}  "+to_str(self.a_state_vecs)+
            f"  state shift  {self.a_state_shift.shape} "+to_str(self.a_state_shift)+
            f"  output vecs  {self.a_output_vecs.shape} "+to_str(self.a_output_vecs)+
             "\n" if (self.a_output_vecs.size > 0) else "") +
            (" model\n"+
            f"  input dimension  {self.config.mdn}{' + '+str(self.config.mde) if self.config.mde > 0 else ''}{' + '+str(self.config.ado) if self.config.ado > 0 else ''}\n"+
            f"  output dimension {self.config.mdo}\n"+
            f"  state dimension  {self.config.mds}\n"+
            f"  number of states {self.config.mns}\n"+
           (f"  embedding dimension  {self.config.mde}\n"+
            f"  number of embeddings {self.config.mne}\n"
             if self.config.mne > 0 else "")+
            f"  embeddings   {self.m_embeddings.shape}  "+to_str(self.m_embeddings)+
            f"  input vecs   {self.m_input_vecs.shape}  "+to_str(self.m_input_vecs)+
            f"  input shift  {self.m_input_shift.shape} "+to_str(self.m_input_shift)+
            f"  state vecs   {self.m_state_vecs.shape}  "+to_str(self.m_state_vecs)+
            f"  state shift  {self.m_state_shift.shape} "+to_str(self.m_state_shift)+
            f"  output vecs  {self.m_output_vecs.shape} "+to_str(self.m_output_vecs)
             if (self.m_output_vecs.size > 0) else "")
        ) + (self._timer_summary_string() if (self.show_times or show_times) else "")


    # Generate the string containing all the configuration information for this model.
    def config_str(self):
        s = ""
        max_n_len = max(map(len,(n for (n,t) in self.config._fields_)))
        max_t_len = max(map(len,(str(t).split("'")[1].split('.')[1]
                                 for (n,t) in self.config._fields_)))
        for (n,t) in self.config._fields_:
            t = str(t).split("'")[1].split('.')[1]
            s += f"  {str(t):{max_t_len}s}  {n:{max_n_len}s}  =  {getattr(self.config,n)}\n"
        return s



# Holder for the model and its work space (with named attributes).
class Details(dict):
    def __init__(self, config, steps=10, ydi=0, ywd=0,
                 model=None, rwork=None, iwork=None, lwork=None,
                 agg_iterators_in=None, record=None, yw=None, yi=None):
        import numpy as np
        self.config = config
        self.steps = steps
        # Generic allocations and objects.
        ftype = dict(order="F", dtype="float32")
        itype = dict(order="F", dtype="int32")
        ltype = dict(order="F", dtype="int64")
        if (model is None):
            model = np.ones(config.total_size, **ftype)
        if (rwork is None):
            rwork = np.ones(config.rwork_size, **ftype)  # beware of allocation, heap vs stack
        if (iwork is None):
            iwork = np.ones(config.iwork_size, **itype)
        if (lwork is None):
            lwork = np.ones(config.lwork_size, **ltype)
        if (agg_iterators_in is None):
            agg_iterators_in = np.ones((6, config.nmt if config.ado > 0 else 0), **ltype)
        if (record is None):
            record = np.zeros((6,steps), **ftype)
        if (yw is None):
            yw = np.zeros((ywd, config.nms), **ftype)
        if (yi is None):
            yi = np.zeros((ydi, config.nms), **ltype)
        # Store source memory allocations internally.
        self.model = model
        self.rwork = rwork
        self.iwork = iwork
        self.lwork = lwork
        # Declare all the special attributes.
        self.update(dict(
            # Model.
            a_embeddings  = model[config.asev-1:config.aeev].reshape(config.ade, config.ane, order="F"),
            a_input_vecs  = model[config.asiv-1:config.aeiv].reshape(config.adi, config.ads, order="F"),
            a_input_shift = model[config.asis-1:config.aeis].reshape(config.ads, order="F"),
            a_state_vecs  = model[config.assv-1:config.aesv].reshape(config.ads, config.ads, max(0,config.ans-1), order="F"),
            a_state_shift = model[config.asss-1:config.aess].reshape(config.ads, max(0,config.ans-1), order="F"),
            a_output_vecs = model[config.asov-1:config.aeov].reshape(config.adso, config.ado+1, order="F"),
            m_embeddings  = model[config.msev-1:config.meev].reshape(config.mde, config.mne, order="F"),
            m_input_vecs  = model[config.msiv-1:config.meiv].reshape(config.mdi, config.mds, order="F"),
            m_input_shift = model[config.msis-1:config.meis].reshape(config.mds, order="F"),
            m_state_vecs  = model[config.mssv-1:config.mesv].reshape(config.mds, config.mds, max(0,config.mns-1), order="F"),
            m_state_shift = model[config.msss-1:config.mess].reshape(config.mds, max(0,config.mns-1), order="F"),
            m_output_vecs = model[config.msov-1:config.meov].reshape(config.mdso, config.mdo, order="F"),
            ax_shift = model[config.aiss-1:config.aise].reshape(config.adn, order="F"),
            ax_rescale = model[config.aims-1:config.aime].reshape(config.adn, config.adn, order="F"),
            ay_shift = model[config.aoss-1:config.aose].reshape(config.ado, order="F"),
            x_shift = model[config.miss-1:config.mise].reshape(config.mdn, order="F"),
            x_rescale = model[config.mims-1:config.mime].reshape(config.mdn, config.mdn, order="F"),
            y_shift = model[config.moss-1:config.mose].reshape(config.do, order="F"),
            y_rescale = model[config.moms-1:config.mome].reshape(config.do, config.do, order="F"),
            # Real work space.
            model_grad = rwork[config.smg-1:config.emg].reshape(config.num_vars, config.num_threads, order="F"),
            model_grad_mean = rwork[config.smgm-1:config.emgm].reshape(config.num_vars, order="F"),
            model_grad_curv = rwork[config.smgc-1:config.emgc].reshape(config.num_vars, order="F"),
            best_model = rwork[config.sbm-1:config.ebm].reshape(config.total_size, order="F"),
            ax = rwork[config.saxb-1:config.eaxb].reshape(config.adi, config.na, order="F"),
            a_emb_temp = rwork[config.saet-1:config.eaet].reshape(config.ade, config.ane, config.num_threads, order="F"),
            a_states = rwork[config.saxs-1:config.eaxs].reshape(config.na, config.ads, config.ans+1, order="F"),
            a_grads = rwork[config.saxg-1:config.eaxg].reshape(config.na, config.ads, config.ans+1, order="F"),
            ay = rwork[config.say-1:config.eay].reshape(config.na, config.ado+1, order="F"),
            ay_gradient = rwork[config.sayg-1:config.eayg].reshape(config.na, config.ado+1, order="F"),
            x = rwork[config.smxb-1:config.emxb].reshape(config.mdi, config.nms, order="F"),
            m_emb_temp = rwork[config.smet-1:config.emet].reshape(config.mde, config.mne, config.num_threads, order="F"),
            m_states = rwork[config.smxs-1:config.emxs].reshape(config.nms, config.mds, config.mns+1, order="F"),
            m_grads = rwork[config.smxg-1:config.emxg].reshape(config.nms, config.mds, config.mns+1, order="F"),
            y = rwork[config.smyb-1:config.emyb].reshape(config.do, config.nms, order="F"),
            y_gradient = rwork[config.syg-1:config.eyg].reshape(config.do, config.nms, order="F"),
            axi_shift = rwork[config.saxis-1:config.eaxis].reshape(config.ade, order="F"),
            axi_rescale = rwork[config.saxir-1:config.eaxir].reshape(config.ade, config.ade, order="F"),
            xi_shift = rwork[config.smxis-1:config.emxis].reshape(config.mde, order="F"),
            xi_rescale = rwork[config.smxir-1:config.emxir].reshape(config.mde, config.mde, order="F"),
            a_lengths = rwork[config.sal-1:config.eal].reshape(config.ads, config.num_threads, order="F"),
            m_lengths = rwork[config.sml-1:config.eml].reshape(config.mds, config.num_threads, order="F"),
            a_state_temp = rwork[config.sast-1:config.east].reshape(config.na, config.ads, order="F"),
            m_state_temp = rwork[config.smst-1:config.emst].reshape(config.nms, config.mds, order="F"),
            # Integer work space.
            axi = lwork[config.saxi-1:config.eaxi].reshape(-1, config.na, order="F") if (config.saxi <= config.eaxi) else np.zeros((0,config.na), **ltype),
            sizes = lwork[config.ssb-1:config.esb].reshape(config.nm, order="F") if (config.ssb <= config.esb) else np.zeros(0, **ltype),
            xi = lwork[config.smxi-1:config.emxi].reshape(-1, config.nms, order="F") if (config.smxi <= config.emxi) else np.zeros((0,config.nms), **ltype),
            a_order = iwork[config.sao-1:config.eao].reshape(config.ads, config.num_threads, order="F"),
            m_order = iwork[config.smo-1:config.emo].reshape(config.mds, config.num_threads, order="F"),
            update_indices = lwork[config.sui-1:config.eui].reshape(config.num_vars, order="F"),
            # External space.
            yi = yi,
            yw = yw,
            record = record,
            agg_iterators_in = agg_iterators_in,
        ))

    def __getattr__(self, *args, **kwargs):
        return self.__getitem__(*args, **kwargs)
    def __setattr__(self, *args, **kwargs):
        return self.__setitem__(*args, **kwargs)
    def __str__(self):
        assert False
        return \
             "Details:\n"+\
            f"  model:   {self.model.dtype.name:8s} {self.model.shape}\n"+\
            f"  rwork:   {self.rwork.dtype.name:8s} {self.rwork.shape}\n"+\
            f"  iwork:   {self.iwork.dtype.name:8s} {self.iwork.shape}\n"+\
            f"  lwork:   {self.lwork.dtype.name:8s} {self.lwork.shape}\n"+\
            f"  a_embeddings:     {self.a_embeddings.shape}\n"+\
            f"  a_input_vecs:     {self.a_input_vecs.shape}\n"+\
            f"  a_input_shift:    {self.a_input_shift.shape}\n"+\
            f"  a_state_vecs:     {self.a_state_vecs.shape}\n"+\
            f"  a_state_shift:    {self.a_state_shift.shape}\n"+\
            f"  a_output_vecs:    {self.a_output_vecs.shape}\n"+\
            f"  m_embeddings:     {self.m_embeddings.shape}\n"+\
            f"  m_input_vecs:     {self.m_input_vecs.shape}\n"+\
            f"  m_input_shift:    {self.m_input_shift.shape}\n"+\
            f"  m_state_vecs:     {self.m_state_vecs.shape}\n"+\
            f"  m_state_shift:    {self.m_state_shift.shape}\n"+\
            f"  m_output_vecs:    {self.m_output_vecs.shape}\n"+\
            f"  ax_shift:         {self.ax_shift.shape}\n"+\
            f"  ax_rescale:       {self.ax_rescale.shape}\n"+\
            f"  ay_shift:         {self.ay_shift.shape}\n"+\
            f"  x_shift:          {getattr(self.x_shift, 'shape', None)}\n"+\
            f"  x_rescale:        {self.x_rescale.shape}\n"+\
            f"  y_shift:          {self.y_shift.shape}\n"+\
            f"  y_rescale:        {self.y_rescale.shape}\n"+\
            f"  model_grad:       {self.model_grad.shape}\n"+\
            f"  model_grad_mean:  {self.model_grad_mean.shape}\n"+\
            f"  model_grad_curv:  {self.model_grad_curv.shape}\n"+\
            f"  best_model:       {self.best_model.shape}\n"+\
            f"  ax:               {self.ax.shape}\n"+\
            f"  a_emb_temp:       {self.a_emb_temp.shape}\n"+\
            f"  a_states:         {self.a_states.shape}\n"+\
            f"  a_grads:          {self.a_grads.shape}\n"+\
            f"  ay:               {self.ay.shape}\n"+\
            f"  ay_gradient:      {self.ay_gradient.shape}\n"+\
            f"  x:                {self.x.shape}\n"+\
            f"  m_emb_temp:       {self.m_emb_temp.shape}\n"+\
            f"  m_states:         {self.m_states.shape}\n"+\
            f"  m_grads:          {self.m_grads.shape}\n"+\
            f"  y:                {self.y.shape}\n"+\
            f"  y_gradient:       {self.y_gradient.shape}\n"+\
            f"  axi_shift:        {self.axi_shift.shape}\n"+\
            f"  axi_rescale:      {self.axi_rescale.shape}\n"+\
            f"  xi_shift:         {self.xi_shift.shape}\n"+\
            f"  xi_rescale:       {self.xi_rescale.shape}\n"+\
            f"  a_lengths:        {self.a_lengths.shape}\n"+\
            f"  m_lengths:        {self.m_lengths.shape}\n"+\
            f"  a_state_temp:     {self.a_state_temp.shape}\n"+\
            f"  m_state_temp:     {self.m_state_temp.shape}\n"+\
            f"  axi:              {self.axi.shape}\n"+\
            f"  sizes:            {self.sizes.shape}\n"+\
            f"  xi:               {self.xi.shape}\n"+\
            f"  a_order:          {self.a_order.shape}\n"+\
            f"  m_order:          {self.m_order.shape}\n"+\
            f"  update_indices:   {self.update_indices.shape}\n"+\
            f"  yi:               {self.yi.shape}\n"+\
            f"  yw:               {self.yw.shape}\n"+\
            f"  record:           {self.record.shape}\n"+\
            f"  agg_iterators_in: {self.agg_iterators_in.shape}\n"
